import os
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Tuple

from bizyengine.core.common.utils import load_config_file, recursive_extract_models


@dataclass
class ModelRule:
    mode_type: str
    base_model: str
    describe: str
    score: int
    route: str
    class_type: str
    inputs: dict


@dataclass
class TaskApi:
    task_result_endpoint: str


class ModelRuleManager:
    def __init__(self, model_rules: list[dict]):
        self.model_rules = model_rules
        self.validate()
        self.gen_model_rule_index_mapping()

    def gen_model_rule_index_mapping(self):
        self.model_rule_index_mapping = defaultdict(list)

        for idx_1, rule in enumerate(self.model_rules):
            for idx_2, node in enumerate(rule["nodes"]):
                self.model_rule_index_mapping[node["class_type"]].append((idx_1, idx_2))

    def validate(self):
        for rule in self.model_rules:
            self._validate_rule(rule)

    def _validate_rule(self, rule: dict):
        if "mode_type" not in rule:
            raise ValueError("mode_type is required")
        if "base_model" not in rule:
            raise ValueError("base_model is required")
        if "route" not in rule:
            raise ValueError("route is required")
        if "nodes" not in rule:
            raise ValueError("nodes is required")

    def find_rule_indexes(self, class_type: str) -> List[Tuple[int, int]]:
        return self.model_rule_index_mapping[class_type]

    def find_rules(self, class_type: str) -> List[ModelRule]:
        rule_indexes = self.find_rule_indexes(class_type)
        return [
            ModelRule(
                mode_type=self.model_rules[idx_1]["mode_type"],
                base_model=self.model_rules[idx_1]["base_model"],
                describe=self.model_rules[idx_1]["describe"],
                score=self.model_rules[idx_1]["score"],
                route=self.model_rules[idx_1]["route"],
                class_type=class_type,
                inputs=self.model_rules[idx_1]["nodes"][idx_2].get("inputs", {}),
            )
            for idx_1, idx_2 in rule_indexes
        ]


class ModelPathManager:
    def __init__(self, config_path: str):
        model_paths = {}
        for folder_name, v in load_config_file(config_path).items():
            model_paths[folder_name] = recursive_extract_models(v)
        self.model_paths = model_paths

    def get_filenames(self, folder_name: str) -> List[str]:
        return self.model_paths.get(folder_name, [])


class ConfigManager:
    def __init__(self, model_path_config: str, model_rule_config: str):
        self.model_path_manager = ModelPathManager(config_path=model_path_config)
        self.model_rule_config = load_config_file(model_rule_config)
        self.model_rules = ModelRuleManager(
            model_rules=self.model_rule_config["model_rules"]
        )

    def get_filenames(self, folder_name: str) -> List[str]:
        return self.model_path_manager.get_filenames(folder_name)

    def get_rules(self, class_type: str) -> List[ModelRule]:
        if class_type.startswith("BizyAir_"):
            class_type = class_type[8:]
        return self.model_rules.find_rules(class_type)

    def get_model_version_id_prefix(self):
        return self.model_rule_config["model_version_config"]["model_version_id_prefix"]

    def get_cache_config(self):
        return self.model_rule_config.get("cache_config", {})

    def get_task_api(self):
        return TaskApi(**self.model_rule_config["task_api"])


model_path_config = os.path.join(os.path.dirname(__file__), "models.json")
model_rule_config = os.path.join(os.path.dirname(__file__), "models.yaml")
config_manager = ConfigManager(
    model_path_config=model_path_config, model_rule_config=model_rule_config
)



from bizyengine.core import BizyAirBaseNode, data_types
from bizyengine.core.path_utils import path_manager as folder_paths

# PuLID Flux datatypes
PULIDFLUX = "BIZYAIR_PULIDFLUX"
EVA_CLIP = "BIZYAIR_EVA_CLIP"
FACEANALYSIS = "BIZYAIR_FACEANALYSIS"


class PulidFluxModelLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {"required": {"pulid_file": (folder_paths.get_filename_list("pulid"),)}}
        # return {
        #     "required": {
        #         "pulid_file": (
        #             [
        #                 "to choose",
        #             ],
        #         ),
        #         "model_version_id": (
        #             "STRING",
        #             {
        #                 "default": "",
        #             },
        #         ),
        #     }
        # }

    RETURN_TYPES = (PULIDFLUX,)
    RETURN_NAMES = ("pulid_flux",)
    # FUNCTION = "load_model"
    CATEGORY = "pulid"
    NODE_DISPLAY_NAME = "Load PuLID Flux Model"


class PulidFluxInsightFaceLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "provider": (
                    [
                        "CUDA",
                    ],
                ),
            },
        }

    RETURN_TYPES = (FACEANALYSIS,)
    RETURN_NAMES = ("face_analysis",)
    # FUNCTION = "load_insightface"
    CATEGORY = "pulid"
    NODE_DISPLAY_NAME = "Load InsightFace (PuLID Flux)"


class PulidFluxEvaClipLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {},
        }

    RETURN_TYPES = (EVA_CLIP,)
    RETURN_NAMES = ("eva_clip",)
    # FUNCTION = "load_eva_clip"
    CATEGORY = "pulid"
    NODE_DISPLAY_NAME = "Load Eva Clip (PuLID Flux)"


class ApplyPulidFlux(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (data_types.MODEL,),
                "pulid_flux": (PULIDFLUX,),
                "eva_clip": (EVA_CLIP,),
                "face_analysis": (FACEANALYSIS,),
                "image": ("IMAGE",),
                "weight": (
                    "FLOAT",
                    {"default": 1.0, "min": -1.0, "max": 5.0, "step": 0.05},
                ),
                "start_at": (
                    "FLOAT",
                    {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
                "end_at": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
            },
            "optional": {
                "attn_mask": ("MASK",),
            },
            "hidden": {"unique_id": "UNIQUE_ID"},
        }

    RETURN_TYPES = (data_types.MODEL,)
    # FUNCTION = "apply_pulid_flux"
    CATEGORY = "pulid"
    NODE_DISPLAY_NAME = "Apply PuLID Flux"



import logging
import os
import uuid
from pathlib import Path

import bizyengine.core.common
import server
from aiohttp import web

# html_file_path = Path(os.path.dirname(os.path.abspath(__file__))) / "set_api_key.html"
# with open(html_file_path, "r", encoding="utf-8") as htmlfile:
#     set_api_key_html = htmlfile.read()


# @server.PromptServer.instance.routes.get("/bizyair/set-api-key")
# async def set_api_key_page(request):
#     return web.Response(text=set_api_key_html, content_type="text/html")


@server.PromptServer.instance.routes.post("/bizyair/set_api_key")
async def set_api_key(request):
    try:
        data = await request.post()
        api_key = data.get("api_key")
        if api_key:
            if not bizyengine.core.common.set_api_key(api_key, True):
                error_msg = "Wrong API key provided, please refer to cloud.siliconflow.cn to get the key"
                logging.error("set_api_key:", error_msg)
                return web.Response(
                    text=error_msg,
                    status=400,
                )
            logging.info("Set api key sucessfully.")
            return web.Response(text="ok")
        else:
            error_msg = "No API key provided, please refer to cloud.siliconflow.cn to get the key"
            logging.error("set_api_key:", error_msg)
            return web.Response(
                text=error_msg,
                status=400,
            )
    except Exception as e:
        logging.error(f"set api key error: {str(e)}")
        return web.Response(text=str(e), status=500)


@server.PromptServer.instance.routes.get("/bizyair/get_api_key")
async def get_api_key(request):
    logging.debug("auth.py get_api_key called")
    try:
        if bizyengine.core.common.get_api_key():
            return web.Response(text="Key has been loaded from the api_key.ini file")
        else:
            logging.debug("getting api key from cookie")
            api_key = request.cookies.get("api_key")
            if not api_key:
                logging.error("No api key found in cookies")
                return web.Response(
                    text="No api key found in cookies, please refer to cloud.siliconflow.cn to get the key",
                    status=404,
                )
            if bizyengine.core.common.set_api_key(api_key):
                return web.Response(text="Key has been loaded from the cookies")
            logging.error("cannot set api key")
            return web.Response(text="Cannot set api key", status=500)

    except Exception as e:
        return web.Response(text=str(e), status=500)


NODE_CLASS_MAPPINGS = {}
NODE_DISPLAY_NAME_MAPPINGS = {}



from aiohttp import web

from .errno import ErrorNo, errnos
from .profile import user_profile


def JsonResponse(http_status_code, data):
    return web.json_response(
        data,
        status=http_status_code,
        content_type="application/json",
    )


def OKResponse(data):
    message = "成功" if user_profile.getLang() == "zh" else "success"
    return JsonResponse(200, {"message": message, "code": errnos.OK.code, "data": data})


def ErrResponse(err: ErrorNo):
    err_msg = err.messages.get(user_profile.getLang())
    if not err_msg:
        err.messages["zh"]

    return JsonResponse(
        err.http_status_code,
        {"message": err_msg, "code": err.code, "data": err.data},
    )



import asyncio
import json

import aiohttp
from aiohttp import web
from bizyengine.core.common.env_var import BIZYAIR_SERVER_ADDRESS
from bizyengine.core.image_utils import decode_data, encode_comfy_image, encode_data
from server import PromptServer

from .utils import (
    decode_and_deserialize,
    get_api_key,
    get_llm_response,
    get_vlm_response,
    send_post_request,
    serialize_and_encode,
)


async def fetch_all_models(api_key):
    url = f"{BIZYAIR_SERVER_ADDRESS}/llm/models"
    headers = {"accept": "application/json", "authorization": f"Bearer {api_key}"}
    params = {"type": "text", "sub_type": "chat"}

    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url, headers=headers, params=params, timeout=10
            ) as response:
                if response.status == 200:
                    data = await response.json()
                    all_models = [model["id"] for model in data["data"]]
                    return all_models
                else:
                    print(f"Error fetching models: HTTP Status {response.status}")
                    return []
    except aiohttp.ClientError as e:
        print(f"Error fetching models: {e}")
        return []
    except asyncio.exceptions.TimeoutError:
        print("Request to fetch models timed out")
        return []


@PromptServer.instance.routes.post("/bizyair/get_silicon_cloud_llm_models")
async def get_silicon_cloud_llm_models_endpoint(request):
    data = await request.json()
    api_key = data.get("api_key", get_api_key())
    all_models = await fetch_all_models(api_key)
    llm_models = [model for model in all_models if "vl" not in model.lower()]
    llm_models.append("No LLM Enhancement")
    return web.json_response(llm_models)


@PromptServer.instance.routes.post("/bizyair/get_silicon_cloud_vlm_models")
async def get_silicon_cloud_vlm_models_endpoint(request):
    data = await request.json()
    api_key = data.get("api_key", get_api_key())
    all_models = await fetch_all_models(api_key)
    vlm_models = [model for model in all_models if "vl" in model.lower()]
    vlm_models.append("No VLM Enhancement")
    return web.json_response(vlm_models)


class SiliconCloudLLMAPI:
    def __init__(self):
        pass

    @classmethod
    def INPUT_TYPES(s):
        default_system_prompt = """你是一个 stable diffusion prompt 专家，为我生成适用于 Stable Diffusion 模型的prompt。 我给你相关的单词，你帮我扩写为适合 Stable Diffusion 文生图的 prompt。要求： 1. 英文输出 2. 除了 prompt 外，不要输出任何其它的信息 """
        return {
            "required": {
                "model": ((), {}),
                "system_prompt": (
                    "STRING",
                    {
                        "default": default_system_prompt,
                        "multiline": True,
                        "dynamicPrompts": True,
                    },
                ),
                "user_prompt": (
                    "STRING",
                    {
                        "default": "小猫，梵高风格",
                        "multiline": True,
                        "dynamicPrompts": True,
                    },
                ),
                "max_tokens": ("INT", {"default": 512, "min": 100, "max": 1e5}),
                "temperature": (
                    "FLOAT",
                    {"default": 0.7, "min": 0.0, "max": 2.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = ("STRING",)
    FUNCTION = "get_llm_model_response"
    OUTPUT_NODE = False
    CATEGORY = "☁️BizyAir/AI Assistants"

    def get_llm_model_response(
        self, model, system_prompt, user_prompt, max_tokens, temperature
    ):
        if model == "No LLM Enhancement":
            return {"ui": {"text": (user_prompt,)}, "result": (user_prompt,)}
        response = get_llm_response(
            model,
            system_prompt,
            user_prompt,
            max_tokens,
            temperature,
        )
        ret = json.loads(response)
        text = ret["choices"][0]["message"]["content"]
        return (text,)  # if update ui:  {"ui": {"text": (text,)}, "result": (text,)}


class SiliconCloudVLMAPI:
    def __init__(self):
        pass

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": ((), {}),
                "system_prompt": (
                    "STRING",
                    {
                        "default": "你是一个能分析图像的AI助手。请仔细观察图像，并根据用户的问题提供详细、准确的描述。",
                        "multiline": True,
                    },
                ),
                "user_prompt": (
                    "STRING",
                    {
                        "default": "请描述这张图片的内容，并指出任何有趣或不寻常的细节。",
                        "multiline": True,
                    },
                ),
                "images": ("IMAGE",),
                "max_tokens": ("INT", {"default": 512, "min": 100, "max": 1e5}),
                "temperature": (
                    "FLOAT",
                    {"default": 0.7, "min": 0.0, "max": 2.0, "step": 0.01},
                ),
                "detail": (["auto", "low", "high"], {"default": "auto"}),
            }
        }

    RETURN_TYPES = ("STRING",)
    FUNCTION = "get_vlm_model_response"
    OUTPUT_NODE = False
    CATEGORY = "☁️BizyAir/AI Assistants"

    def get_vlm_model_response(
        self, model, system_prompt, user_prompt, images, max_tokens, temperature, detail
    ):
        if model == "No VLM Enhancement":
            return (user_prompt,)

        # 使用 encode_comfy_image 函数编码图像批次
        encoded_images_json = encode_comfy_image(
            images, image_format="WEBP", lossless=True
        )
        encoded_images_dict = json.loads(encoded_images_json)

        # 提取所有编码后的图像
        base64_images = list(encoded_images_dict.values())

        response = get_vlm_response(
            model,
            system_prompt,
            user_prompt,
            base64_images,
            max_tokens,
            temperature,
            detail,
        )
        ret = json.loads(response)
        text = ret["choices"][0]["message"]["content"]
        return (text,)


class BizyAirJoyCaption:
    # refer to: https://huggingface.co/spaces/fancyfeast/joy-caption-pre-alpha
    API_URL = f"{BIZYAIR_SERVER_ADDRESS}/supernode/joycaption2"

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "image": ("IMAGE",),
                "do_sample": (["enable", "disable"],),
                "temperature": (
                    "FLOAT",
                    {
                        "default": 0.5,
                        "min": 0.0,
                        "max": 2.0,
                        "step": 0.01,
                        "round": 0.001,
                        "display": "number",
                    },
                ),
                "max_tokens": (
                    "INT",
                    {
                        "default": 256,
                        "min": 16,
                        "max": 512,
                        "step": 16,
                        "display": "number",
                    },
                ),
            }
        }

    RETURN_TYPES = ("STRING",)
    FUNCTION = "joycaption"

    CATEGORY = "☁️BizyAir/AI Assistants"

    def joycaption(self, image, do_sample, temperature, max_tokens):
        API_KEY = get_api_key()
        SIZE_LIMIT = 1536
        # device = image.device
        _, w, h, c = image.shape
        assert (
            w <= SIZE_LIMIT and h <= SIZE_LIMIT
        ), f"width and height must be less than {SIZE_LIMIT}x{SIZE_LIMIT}, but got {w} and {h}"

        payload = {
            "image": None,
            "do_sample": do_sample == "enable",
            "temperature": temperature,
            "max_new_tokens": max_tokens,
            "caption_type": "Descriptive",
            "caption_length": "any",
            "extra_options": [],
            "name_input": "",
            "custom_prompt": "A descriptive caption for this image:\n",
        }
        auth = f"Bearer {API_KEY}"
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "authorization": auth,
        }
        input_image = encode_data(image, disable_image_marker=True)
        payload["image"] = input_image

        ret: str = send_post_request(self.API_URL, payload=payload, headers=headers)
        ret = json.loads(ret)

        try:
            if "result" in ret:
                ret = json.loads(ret["result"])
            if ret["type"] == "error":
                raise Exception(ret["message"])
        except Exception as e:
            raise Exception(f"Unexpected response: {ret} {e=}")

        msg = ret["data"]
        if msg["type"] not in (
            "comfyair",
            "bizyair",
        ):
            raise Exception(f"Unexpected response type: {msg}")

        caption = msg["data"]
        return (caption,)


class BizyAirJoyCaption2:
    def __init__(self):
        pass

    # refer to: https://huggingface.co/spaces/fancyfeast/joy-caption-pre-alpha
    API_URL = f"{BIZYAIR_SERVER_ADDRESS}/supernode/joycaption2"

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "image": ("IMAGE",),
                "do_sample": ([True, False],),
                "temperature": (
                    "FLOAT",
                    {
                        "default": 0.5,
                        "min": 0.0,
                        "max": 2.0,
                        "step": 0.01,
                        "round": 0.001,
                        "display": "number",
                    },
                ),
                "max_tokens": (
                    "INT",
                    {
                        "default": 256,
                        "min": 16,
                        "max": 512,
                        "step": 16,
                        "display": "number",
                    },
                ),
                "caption_type": (
                    [
                        "Descriptive",
                        "Descriptive (Informal)",
                        "Training Prompt",
                        "MidJourney",
                        "Booru tag list",
                        "Booru-like tag list",
                        "Art Critic",
                        "Product Listing",
                        "Social Media Post",
                    ],
                ),
                "caption_length": (
                    ["any", "very short", "short", "medium-length", "long", "very long"]
                    + [str(i) for i in range(20, 261, 10)],
                ),
                "extra_options": (
                    "STRING",
                    {
                        "default": "If there is a person/character in the image you must refer to them as {name}.",
                        "tooltip": "Extra options for the model",
                        "multiline": True,
                    },
                ),
                "name_input": (
                    "STRING",
                    {
                        "default": "Jack",
                        "tooltip": "Name input is only used if an Extra Option is selected that requires it.",
                    },
                ),
                "custom_prompt": (
                    "STRING",
                    {
                        "default": "",
                        "multiline": True,
                    },
                ),
            }
        }

    RETURN_TYPES = ("STRING",)
    FUNCTION = "joycaption2"

    CATEGORY = "☁️BizyAir/AI Assistants"

    def joycaption2(
        self,
        image,
        do_sample,
        temperature,
        max_tokens,
        caption_type,
        caption_length,
        extra_options,
        name_input,
        custom_prompt,
    ):
        API_KEY = get_api_key()
        SIZE_LIMIT = 1536
        _, w, h, c = image.shape
        assert (
            w <= SIZE_LIMIT and h <= SIZE_LIMIT
        ), f"width and height must be less than {SIZE_LIMIT}x{SIZE_LIMIT}, but got {w} and {h}"

        payload = {
            "image": None,
            "do_sample": do_sample == True,
            "temperature": temperature,
            "max_new_tokens": max_tokens,
            "caption_type": caption_type,
            "caption_length": caption_length,
            "extra_options": [extra_options],
            "name_input": name_input,
            "custom_prompt": custom_prompt,
        }
        auth = f"Bearer {API_KEY}"
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "authorization": auth,
        }
        input_image = encode_data(image, disable_image_marker=True)
        payload["image"] = input_image

        ret: str = send_post_request(self.API_URL, payload=payload, headers=headers)
        ret = json.loads(ret)

        try:
            if "result" in ret:
                ret = json.loads(ret["result"])
            if ret["type"] == "error":
                raise Exception(ret["message"])
        except Exception as e:
            raise Exception(f"Unexpected response: {ret} {e=}")

        msg = ret["data"]
        if msg["type"] not in (
            "comfyair",
            "bizyair",
        ):
            raise Exception(f"Unexpected response type: {msg}")

        caption = msg["data"]
        return (caption,)


NODE_CLASS_MAPPINGS = {
    "BizyAirSiliconCloudLLMAPI": SiliconCloudLLMAPI,
    "BizyAirSiliconCloudVLMAPI": SiliconCloudVLMAPI,
    "BizyAirJoyCaption": BizyAirJoyCaption,
    "BizyAirJoyCaption2": BizyAirJoyCaption2,
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "BizyAirSiliconCloudLLMAPI": "☁️BizyAir SiliconCloud LLM API",
    "BizyAirSiliconCloudVLMAPI": "☁️BizyAir SiliconCloud VLM API",
    "BizyAirJoyCaption": "☁️BizyAir Joy Caption",
    "BizyAirJoyCaption2": "☁️BizyAir Joy Caption2",
}



from bizyengine.core import BizyAirBaseNode


class JanusModelLoader(BizyAirBaseNode):

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model_name": (
                    ["deepseek-ai/Janus-Pro-7B"],
                ),  # "deepseek-ai/Janus-Pro-1B",
            },
        }

    RETURN_TYPES = ("BIZYAIR_JANUS_MODEL", "BIZYAIR_JANUS_PROCESSOR")
    RETURN_NAMES = ("model", "processor")
    # FUNCTION = "load_model"
    CATEGORY = "Janus-Pro"


class JanusImageUnderstanding(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": ("BIZYAIR_JANUS_MODEL",),
                "processor": ("BIZYAIR_JANUS_PROCESSOR",),
                "image": ("IMAGE",),
                "question": (
                    "STRING",
                    {"multiline": True, "default": "Describe this image in detail."},
                ),
                "seed": (
                    "INT",
                    {"default": 666666666666666, "min": 0, "max": 0xFFFFFFFFFFFFFFFF},
                ),
                "temperature": ("FLOAT", {"default": 0.1, "min": 0.0, "max": 1.0}),
                "top_p": ("FLOAT", {"default": 0.95, "min": 0.0, "max": 1.0}),
                "max_new_tokens": ("INT", {"default": 512, "min": 1, "max": 2048}),
            },
        }

    RETURN_TYPES = ("STRING",)
    RETURN_NAMES = ("text",)
    # FUNCTION = "analyze_image"
    CATEGORY = "Janus-Pro"


class JanusImageGeneration(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": ("BIZYAIR_JANUS_MODEL",),
                "processor": ("BIZYAIR_JANUS_PROCESSOR",),
                "prompt": (
                    "STRING",
                    {"multiline": True, "default": "A beautiful photo of"},
                ),
                "seed": (
                    "INT",
                    {"default": 666666666666666, "min": 0, "max": 0xFFFFFFFFFFFFFFFF},
                ),
                "batch_size": ("INT", {"default": 4, "min": 4, "max": 9}),
                "cfg_weight": (
                    "FLOAT",
                    {"default": 5.0, "min": 1.0, "max": 10.0, "step": 0.5},
                ),
                "temperature": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.1, "max": 2.0, "step": 0.1},
                ),
                "top_p": ("FLOAT", {"default": 0.95, "min": 0.0, "max": 1.0}),
            },
        }

    RETURN_TYPES = ("IMAGE",)
    RETURN_NAMES = ("images",)
    # FUNCTION = "generate_images"
    CATEGORY = "Janus-Pro"



import os
from typing import List

import comfy
from bizyengine.core import BizyAirBaseNode, BizyAirNodeIO, create_node_data, data_types
from bizyengine.core.configs.conf import config_manager
from bizyengine.core.path_utils import path_manager as folder_paths

LOGO = "☁️"
PREFIX = f"{LOGO}BizyAir"

MAX_RESOLUTION = 16384  # https://github.com/comfyanonymous/ComfyUI/blob/7390ff3b1ec2e15017ba4a52d6eaabc4aa4636e3/nodes.py#L45


class ProgressCallback:
    def __init__(self, total=None) -> None:
        comfy.model_management.throw_exception_if_processing_interrupted()
        self.pbar = comfy.utils.ProgressBar(None)

    def __call__(self, value, total=None, preview=None):
        self.pbar.update_absolute(value, total, preview)


class BizyAir_KSampler(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (data_types.MODEL,),
                "seed": (
                    "INT",
                    {"default": 0, "min": 0, "max": 0xFFFFFFFFFFFFFFFF},
                ),
                "steps": ("INT", {"default": 20, "min": 1, "max": 50}),
                "cfg": (
                    "FLOAT",
                    {
                        "default": 8.0,
                        "min": 0.0,
                        "max": 100.0,
                        "step": 0.1,
                        "round": 0.01,
                    },
                ),
                "sampler_name": (comfy.samplers.KSampler.SAMPLERS,),
                "scheduler": (comfy.samplers.KSampler.SCHEDULERS,),
                "positive": (data_types.CONDITIONING,),
                "negative": (data_types.CONDITIONING,),
                "latent_image": ("LATENT",),
                "denoise": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = ("LATENT",)
    FUNCTION = "sample"
    RETURN_NAMES = (f"LATENT",)
    CATEGORY = f"{PREFIX}/sampling"

    def sample(
        self,
        model,
        seed,
        steps,
        cfg,
        sampler_name,
        scheduler,
        positive,
        negative,
        latent_image,
        denoise=1,
    ):
        new_model: BizyAirNodeIO = model.copy(self.assigned_id)
        new_model.add_node_data(
            class_type="KSampler",
            inputs={
                "model": model,
                "seed": seed,
                "steps": steps,
                "cfg": cfg,
                "sampler_name": sampler_name,
                "scheduler": scheduler,
                "positive": positive,
                "negative": negative,
                "latent_image": latent_image,
                "denoise": denoise,
            },
            outputs={"slot_index": 0},
        )
        progress_callback = ProgressCallback()
        return new_model.send_request(progress_callback=progress_callback)


class KSamplerAdvanced(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (data_types.MODEL,),
                "add_noise": (["enable", "disable"],),
                "noise_seed": (
                    "INT",
                    {"default": 0, "min": 0, "max": 0xFFFFFFFFFFFFFFFF},
                ),
                "steps": ("INT", {"default": 20, "min": 1, "max": 10000}),
                "cfg": (
                    "FLOAT",
                    {
                        "default": 8.0,
                        "min": 0.0,
                        "max": 100.0,
                        "step": 0.1,
                        "round": 0.01,
                    },
                ),
                "sampler_name": (comfy.samplers.KSampler.SAMPLERS,),
                "scheduler": (comfy.samplers.KSampler.SCHEDULERS,),
                "positive": (data_types.CONDITIONING,),
                "negative": (data_types.CONDITIONING,),
                "latent_image": ("LATENT",),
                "start_at_step": ("INT", {"default": 0, "min": 0, "max": 10000}),
                "end_at_step": ("INT", {"default": 10000, "min": 0, "max": 10000}),
                "return_with_leftover_noise": (["disable", "enable"],),
            }
        }

    RETURN_TYPES = ("LATENT",)
    FUNCTION = "sample"

    CATEGORY = "sampling"

    def sample(self, model, **kwargs):
        new_model: BizyAirNodeIO = model.copy(self.assigned_id)
        kwargs["model"] = model
        new_model.add_node_data(class_type="KSamplerAdvanced", inputs=kwargs)
        progress_callback = ProgressCallback()
        return new_model.send_request(progress_callback=progress_callback)


class BizyAir_CheckpointLoaderSimple(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "ckpt_name": (
                    [
                        "to choose",
                    ],
                ),
                "model_version_id": (
                    "STRING",
                    {
                        "default": "",
                    },
                ),
            }
        }

    RETURN_TYPES = (data_types.MODEL, data_types.CLIP, data_types.VAE)
    FUNCTION = "load_checkpoint"
    CATEGORY = f"{PREFIX}/loaders"
    RETURN_NAMES = (
        f"model",
        f"clip",
        f"vae",
    )

    @classmethod
    def VALIDATE_INPUTS(cls, ckpt_name):
        # TODO
        import warnings

        warnings.warn(message=f"TODO fix {cls}VALIDATE_INPUTS")
        if ckpt_name == "" or ckpt_name is None:
            return False
        return True

    def load_checkpoint(self, ckpt_name, model_version_id=""):
        if model_version_id != "":
            # use model version id as lora name
            ckpt_name = (
                f"{config_manager.get_model_version_id_prefix()}{model_version_id}"
            )
        node_datas = [
            create_node_data(
                class_type="CheckpointLoaderSimple",
                inputs={"ckpt_name": ckpt_name},
                outputs={"slot_index": slot_index},
            )
            for slot_index in range(3)
        ]
        config_file = folder_paths.guess_config(ckpt_name=ckpt_name)
        assigned_id = self.assigned_id
        outs = [
            BizyAirNodeIO(
                assigned_id,
                {assigned_id: data},
                config_file=config_file,
            )
            for data in node_datas
        ]

        return (
            outs[0],
            outs[1],
            outs[2],
        )


class BizyAir_CLIPTextEncode(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "text": (
                    "STRING",
                    {"multiline": True, "dynamicPrompts": True},
                ),
                "clip": (data_types.CLIP,),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    RETURN_NAMES = ("CONDITIONING",)
    FUNCTION = "encode"

    CATEGORY = f"{PREFIX}/conditioning"

    def encode(self, clip, text):
        new_clip: BizyAirNodeIO = clip.copy(self.assigned_id)

        new_clip.add_node_data(
            class_type="CLIPTextEncode",
            inputs={
                "text": text,
                "clip": clip,
            },
            outputs={"slot_index": 0},
        )
        return (new_clip,)


class BizyAir_VAEDecode(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {"required": {"samples": ("LATENT",), "vae": (data_types.VAE,)}}

    RETURN_TYPES = ("IMAGE",)
    RETURN_NAMES = (f"IMAGE",)
    FUNCTION = "decode"

    CATEGORY = f"{PREFIX}/latent"

    def decode(self, vae, samples):
        new_vae: BizyAirNodeIO = vae.copy(self.assigned_id)
        new_vae.add_node_data(
            class_type="VAEDecode",
            inputs={
                "samples": samples,
                "vae": vae,
            },
            outputs={"slot_index": 0},
        )
        return new_vae.send_request()


class BizyAir_LoraLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (data_types.MODEL,),
                "clip": (data_types.CLIP,),
                "lora_name": (
                    [
                        "to choose",
                    ],
                ),
                "strength_model": (
                    "FLOAT",
                    {"default": 1.0, "min": -100.0, "max": 100.0, "step": 0.01},
                ),
                "strength_clip": (
                    "FLOAT",
                    {"default": 1.0, "min": -100.0, "max": 100.0, "step": 0.01},
                ),
                "model_version_id": (
                    "STRING",
                    {
                        "default": "",
                    },
                ),
            }
        }

    RETURN_TYPES = (data_types.MODEL, data_types.CLIP)
    RETURN_NAMES = ("MODEL", "CLIP")

    FUNCTION = "load_lora"
    CATEGORY = f"{PREFIX}/loaders"

    def load_lora(
        self,
        model,
        clip,
        lora_name,
        strength_model,
        strength_clip,
        model_version_id: str = None,
    ):
        assigned_id = self.assigned_id
        new_model: BizyAirNodeIO = model.copy(assigned_id)
        new_clip: BizyAirNodeIO = clip.copy(assigned_id)
        instances: List[BizyAirNodeIO] = [new_model, new_clip]

        if model_version_id is not None and model_version_id != "":
            # use model version id as lora name
            lora_name = (
                f"{config_manager.get_model_version_id_prefix()}{model_version_id}"
            )

        for slot_index, ins in zip(range(2), instances):
            ins.add_node_data(
                class_type="LoraLoader",
                inputs={
                    "model": model,
                    "clip": clip,
                    "lora_name": lora_name,
                    "strength_model": strength_model,
                    "strength_clip": strength_clip,
                },
                outputs={"slot_index": slot_index},
            )
        return (
            new_model,
            new_clip,
        )

    @classmethod
    def VALIDATE_INPUTS(cls, lora_name):
        if lora_name == "" or lora_name is None:
            return False
        return True


class BizyAir_LoraLoader_Legacy(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (data_types.MODEL,),
                "clip": (data_types.CLIP,),
                "lora_name": (folder_paths.get_filename_list("loras"),),
                "strength_model": (
                    "FLOAT",
                    {"default": 1.0, "min": -100.0, "max": 100.0, "step": 0.01},
                ),
                "strength_clip": (
                    "FLOAT",
                    {"default": 1.0, "min": -100.0, "max": 100.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = (data_types.MODEL, data_types.CLIP)
    RETURN_NAMES = ("MODEL", "CLIP")

    FUNCTION = "load_lora"

    CATEGORY = f"{PREFIX}/loaders"

    def load_lora(self, model, clip, lora_name, strength_model, strength_clip):
        assigned_id = self.assigned_id
        new_model: BizyAirNodeIO = model.copy(assigned_id)
        new_clip: BizyAirNodeIO = clip.copy(assigned_id)
        instances: List[BizyAirNodeIO] = [new_model, new_clip]
        for slot_index, ins in zip(range(2), instances):
            ins.add_node_data(
                class_type="LoraLoader",
                inputs={
                    "model": model,
                    "clip": clip,
                    "lora_name": lora_name,
                    "strength_model": strength_model,
                    "strength_clip": strength_clip,
                },
                outputs={"slot_index": slot_index},
            )
        return (
            new_model,
            new_clip,
        )


class BizyAir_VAEEncode(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {"required": {"pixels": ("IMAGE",), "vae": (data_types.VAE,)}}

    RETURN_TYPES = ("LATENT",)
    RETURN_NAMES = (f"LATENT",)
    FUNCTION = "encode"
    CATEGORY = f"{PREFIX}/latent"

    def encode(self, vae, pixels):
        new_vae: BizyAirNodeIO = vae.copy(self.assigned_id)
        new_vae.add_node_data(
            class_type="VAEEncode",
            inputs={
                "vae": vae,
                "pixels": pixels,
            },
            outputs={"slot_index": 0},
        )
        return new_vae.send_request()


class BizyAir_VAEEncodeForInpaint(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "pixels": ("IMAGE",),
                "vae": (data_types.VAE,),
                "mask": ("MASK",),
                "grow_mask_by": ("INT", {"default": 6, "min": 0, "max": 64, "step": 1}),
            }
        }

    RETURN_TYPES = (f"LATENT",)
    RETURN_NAMES = (f"LATENT",)
    FUNCTION = "encode"
    CATEGORY = f"{PREFIX}/latent/inpaint"

    def encode(self, vae, pixels, mask, grow_mask_by=6):
        new_vae: BizyAirNodeIO = vae.copy(self.assigned_id)
        new_vae.add_node_data(
            class_type="VAEEncodeForInpaint",
            inputs={
                "vae": vae,
                "pixels": pixels,
                "mask": mask,
                "grow_mask_by": grow_mask_by,
            },
            outputs={"slot_index": 0},
        )
        return new_vae.send_request()


class BizyAir_ControlNetLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "control_net_name": (
                    [
                        "to choose",
                    ],
                ),
                "model_version_id": ("STRING", {"default": "", "multiline": False}),
            }
        }

    RETURN_TYPES = (data_types.CONTROL_NET,)
    RETURN_NAMES = ("CONTROL_NET",)
    FUNCTION = "load_controlnet"

    CATEGORY = f"{PREFIX}/loaders"

    @classmethod
    def VALIDATE_INPUTS(cls, control_net_name, model_version_id):
        if control_net_name == "to choose":
            return False
        if model_version_id is not None and model_version_id != "":
            return True
        return True

    def load_controlnet(self, control_net_name, model_version_id):
        if model_version_id is not None and model_version_id != "":
            control_net_name = (
                f"{config_manager.get_model_version_id_prefix()}{model_version_id}"
            )

        node_data = create_node_data(
            class_type="ControlNetLoader",
            inputs={
                "control_net_name": control_net_name,
            },
            outputs={"slot_index": 0},
        )
        assigned_id = self.assigned_id
        node = BizyAirNodeIO(assigned_id, {assigned_id: node_data})
        return (node,)


class BizyAir_ControlNetLoader_Legacy(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "control_net_name": (folder_paths.get_filename_list("controlnet"),)
            }
        }

    RETURN_TYPES = (data_types.CONTROL_NET,)
    RETURN_NAMES = ("CONTROL_NET",)
    FUNCTION = "load_controlnet"

    CATEGORY = f"{PREFIX}/loaders"

    def load_controlnet(self, control_net_name):

        node_data = create_node_data(
            class_type="ControlNetLoader",
            inputs={
                "control_net_name": control_net_name,
            },
            outputs={"slot_index": 0},
        )
        assigned_id = self.assigned_id
        node = BizyAirNodeIO(assigned_id, {assigned_id: node_data})
        return (node,)


class BizyAir_ControlNetApplyAdvanced(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "positive": (data_types.CONDITIONING,),
                "negative": (data_types.CONDITIONING,),
                "control_net": (data_types.CONTROL_NET,),
                "image": ("IMAGE",),
                "strength": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 10.0, "step": 0.01},
                ),
                "start_percent": (
                    "FLOAT",
                    {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
                "end_percent": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING, data_types.CONDITIONING)
    RETURN_NAMES = ("positive", "negative")
    FUNCTION = "apply_controlnet"

    CATEGORY = "conditioning"

    def apply_controlnet(self, **kwargs):
        new_positive = kwargs["positive"].copy(self.assigned_id)
        new_negative = kwargs["negative"].copy(self.assigned_id)
        outs = [
            new_positive,
            new_negative,
        ]
        for slot_index, out in zip(range(0, 2), outs):
            out.add_node_data(
                class_type="ControlNetApplyAdvanced",
                inputs=kwargs,
                outputs={"slot_index": slot_index},
            )
        return outs


class BizyAir_ControlNetApply(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning": (data_types.CONDITIONING,),
                "control_net": (data_types.CONTROL_NET,),
                "image": ("IMAGE",),
                "strength": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 10.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    RETURN_NAMES = ("CONDITIONING",)
    FUNCTION = "apply_controlnet"

    CATEGORY = f"{PREFIX}/conditioning/controlnet"

    def apply_controlnet(
        self, conditioning, control_net: BizyAirNodeIO, image, strength
    ):
        new_cond: BizyAirNodeIO = conditioning.copy(self.assigned_id)
        new_cond.add_node_data(
            class_type="ControlNetApply",
            inputs={
                "conditioning": conditioning,
                "control_net": control_net,
                "image": image,
                "strength": strength,
            },
            outputs={"slot_index": 0},
        )
        return (new_cond,)


class BizyAir_CLIPVisionLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "clip_name": (folder_paths.get_filename_list("clip_vision"),),
                # "clip_name": (
                #     [
                #         "to choose",
                #     ],
                # ),
                # "model_version_id": (
                #     "STRING",
                #     {
                #         "default": "",
                #     },
                # ),
            }
        }

    RETURN_TYPES = ("CLIP_VISION",)
    FUNCTION = "load_clip"

    CATEGORY = "loaders"

    def load_clip(self, clip_name):
        node_data = create_node_data(
            class_type="CLIPVisionLoader",
            inputs={"clip_name": clip_name},
            outputs={"slot_index": 0},
        )
        clip_vision = BizyAirNodeIO(self.assigned_id, {self.assigned_id: node_data})
        return (clip_vision,)

    # @classmethod
    # def VALIDATE_INPUTS(cls, clip_name):
    #     # TODO
    #     import warnings

    #     warnings.warn(message=f"TODO fix {cls}VALIDATE_INPUTS")
    #     if clip_name == "" or clip_name is None:
    #         return False
    #     return True

    # def load_clip(self, clip_name, model_version_id=""):
    #     if model_version_id != "":
    #         # use model version id as lora name
    #         clip_name = (
    #             f"{config_manager.get_model_version_id_prefix()}{model_version_id}"
    #         )
    #     node_data = create_node_data(
    #         class_type="CLIPVisionLoader",
    #         inputs={"clip_name": clip_name},
    #         outputs={"slot_index": 0},
    #     )
    #     clip_vision = BizyAirNodeIO(self.assigned_id, {self.assigned_id: node_data})
    #     return (clip_vision,)


class VAELoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {"required": {"vae_name": (folder_paths.get_filename_list("vae"),)}}
        # return {
        #     "required": {
        #         "vae_name": (
        #             [
        #                 "to choose",
        #             ],
        #         ),
        #         "model_version_id": (
        #             "STRING",
        #             {
        #                 "default": "",
        #             },
        #         ),
        #     }
        # }

    RETURN_TYPES = (data_types.VAE,)
    RETURN_NAMES = ("vae",)
    FUNCTION = "load_vae"

    CATEGORY = "loaders"

    def load_vae(self, vae_name):
        node_data = create_node_data(
            class_type="VAELoader",
            inputs={"vae_name": vae_name},
            outputs={"slot_index": 0},
        )
        vae = BizyAirNodeIO(
            self.assigned_id,
            {self.assigned_id: node_data},
            config_file=folder_paths.guess_config(vae_name=vae_name),
        )
        return (vae,)

    # @classmethod
    # def VALIDATE_INPUTS(cls, vae_name):
    #     # TODO
    #     import warnings

    #     warnings.warn(message=f"TODO fix {cls}VALIDATE_INPUTS")
    #     if vae_name == "" or vae_name is None:
    #         return False
    #     return True

    # def load_vae(self, vae_name, model_version_id):
    #     if model_version_id != "":
    #         # use model version id as lora name
    #         vae_name = (
    #             f"{config_manager.get_model_version_id_prefix()}{model_version_id}"
    #         )
    #     node_data = create_node_data(
    #         class_type="VAELoader",
    #         inputs={"vae_name": vae_name},
    #         outputs={"slot_index": 0},
    #     )
    #     vae = BizyAirNodeIO(
    #         self.assigned_id,
    #         {self.assigned_id: node_data},
    #         config_file=folder_paths.guess_config(vae_name=vae_name),
    #     )
    #     return (vae,)


class UNETLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "unet_name": (folder_paths.get_filename_list("unet"),),
                # "unet_name": (
                #     [
                #         "to choose",
                #     ],
                # ),
                # "model_version_id": (
                #     "STRING",
                #     {
                #         "default": "",
                #     },
                # ),
                "weight_dtype": (["default", "fp8_e4m3fn", "fp8_e5m2"],),
            }
        }

    RETURN_TYPES = (data_types.MODEL,)
    FUNCTION = "load_unet"

    CATEGORY = "advanced/loaders"

    def load_unet(self, unet_name, weight_dtype):
        node_data = create_node_data(
            class_type="UNETLoader",
            inputs={
                "unet_name": unet_name,
                "weight_dtype": weight_dtype,
            },
            outputs={"slot_index": 0},
        )
        model = BizyAirNodeIO(
            self.assigned_id,
            {self.assigned_id: node_data},
            config_file=folder_paths.guess_config(unet_name=unet_name),
        )
        return (model,)

    # @classmethod
    # def VALIDATE_INPUTS(cls, unet_name):
    #     # TODO
    #     import warnings

    #     warnings.warn(message=f"TODO fix {cls}VALIDATE_INPUTS")
    #     if unet_name == "" or unet_name is None:
    #         return False
    #     return True

    # def load_unet(self, unet_name, model_version_id, weight_dtype):
    #     if model_version_id != "":
    #         # use model version id as lora name
    #         unet_name = (
    #             f"{config_manager.get_model_version_id_prefix()}{model_version_id}"
    #         )
    #     node_data = create_node_data(
    #         class_type="UNETLoader",
    #         inputs={
    #             "unet_name": unet_name,
    #             "weight_dtype": weight_dtype,
    #         },
    #         outputs={"slot_index": 0},
    #     )
    #     model = BizyAirNodeIO(
    #         self.assigned_id,
    #         {self.assigned_id: node_data},
    #         config_file=folder_paths.guess_config(unet_name=unet_name),
    #     )
    #     return (model,)


class SamplerCustomAdvanced(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "noise": ("NOISE",),
                "guider": ("GUIDER",),
                "sampler": ("SAMPLER",),
                "sigmas": ("SIGMAS",),
                "latent_image": ("LATENT",),
            }
        }

    RETURN_TYPES = ("LATENT", "LATENT")
    RETURN_NAMES = ("output", "denoised_output")

    # FUNCTION = "sample"

    CATEGORY = "sampling/custom_sampling"

    # def sample(self, **kwargs):
    #     guider: BizyAirNodeIO = kwargs["guider"].copy(self.assigned_id)
    #     guider.add_node_data(
    #         class_type="SamplerCustomAdvanced",
    #         inputs=kwargs,
    #         outputs={"slot_index": 0},
    #     )
    #     return (guider, None)


class BasicGuider(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (data_types.MODEL,),
                "conditioning": (data_types.CONDITIONING,),
            }
        }

    RETURN_TYPES = ("GUIDER",)

    FUNCTION = "get_guider"
    CATEGORY = "sampling/custom_sampling/guiders"

    def get_guider(self, model: BizyAirNodeIO, conditioning):
        new_model = model.copy(self.assigned_id)
        new_model.add_node_data(
            class_type="BasicGuider",
            inputs={
                "model": model,
                "conditioning": conditioning,
            },
            outputs={"slot_index": 0},
        )
        return (new_model,)


class BasicScheduler(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (data_types.MODEL,),
                "scheduler": (comfy.samplers.SCHEDULER_NAMES,),
                "steps": ("INT", {"default": 20, "min": 1, "max": 10000}),
                "denoise": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = ("SIGMAS",)
    CATEGORY = "sampling/custom_sampling/schedulers"

    FUNCTION = "get_sigmas"

    def get_sigmas(self, **kwargs):
        new_model: BizyAirNodeIO = kwargs["model"].copy(self.assigned_id)
        new_model.add_node_data(
            class_type="BasicScheduler",
            inputs=kwargs,
            outputs={"slot_index": 0},
        )
        return (new_model,)


class DualCLIPLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "clip_name1": (folder_paths.get_filename_list("clip"),),
                "clip_name2": (folder_paths.get_filename_list("clip"),),
                # "clip_name1": (
                #     [
                #         "to choose",
                #     ],
                # ),
                # "model_version_id1": (
                #     "STRING",
                #     {
                #         "default": "",
                #     },
                # ),
                # "clip_name2": (
                #     [
                #         "to choose",
                #     ],
                # ),
                # "model_version_id2": (
                #     "STRING",
                #     {
                #         "default": "",
                #     },
                # ),
                "type": (["sdxl", "sd3", "flux"],),
            }
        }

    RETURN_TYPES = (data_types.CLIP,)
    FUNCTION = "load_clip"

    CATEGORY = "advanced/loaders"

    def load_clip(self, clip_name1, clip_name2, type):
        node_data = create_node_data(
            class_type="DualCLIPLoader",
            inputs={
                "clip_name1": clip_name1,
                "clip_name2": clip_name2,
                "type": type,
            },
            outputs={"slot_index": 0},
        )
        model = BizyAirNodeIO(
            self.assigned_id,
            {self.assigned_id: node_data},
            config_file=folder_paths.guess_config(clip_name=clip_name1),
        )
        return (model,)

    # @classmethod
    # def VALIDATE_INPUTS(cls, ckpt_name):
    #     # TODO
    #     import warnings

    #     warnings.warn(message=f"TODO fix {cls}VALIDATE_INPUTS")
    #     if ckpt_name == "" or ckpt_name is None:
    #         return False
    #     return True

    # def load_clip(
    #     self, clip_name1, model_version_id1, clip_name2, model_version_id2, type
    # ):
    #     if model_version_id1 != "":
    #         # use model version id as lora name
    #         clip_name1 = (
    #             f"{config_manager.get_model_version_id_prefix()}{model_version_id1}"
    #         )
    #     if model_version_id2 != "":
    #         # use model version id as lora name
    #         clip_name2 = (
    #             f"{config_manager.get_model_version_id_prefix()}{model_version_id2}"
    #         )
    #     node_data = create_node_data(
    #         class_type="DualCLIPLoader",
    #         inputs={
    #             "clip_name1": clip_name1,
    #             "clip_name2": clip_name2,
    #             "type": type,
    #         },
    #         outputs={"slot_index": 0},
    #     )
    #     model = BizyAirNodeIO(
    #         self.assigned_id,
    #         {self.assigned_id: node_data},
    #         config_file=folder_paths.guess_config(clip_name=clip_name1),
    #     )
    #     return (model,)


class KSamplerSelect(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "sampler_name": (comfy.samplers.SAMPLER_NAMES,),
            }
        }

    RETURN_TYPES = ("SAMPLER",)
    CATEGORY = "sampling/custom_sampling/samplers"

    FUNCTION = "get_sampler"

    def get_sampler(self, **kwargs):
        node_data = create_node_data(
            class_type="KSamplerSelect",
            inputs=kwargs,
            outputs={"slot_index": 0},
        )
        model = BizyAirNodeIO(
            self.assigned_id,
            {self.assigned_id: node_data},
        )
        return (model,)


class RandomNoise(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "noise_seed": (
                    "INT",
                    {"default": 0, "min": 0, "max": 0xFFFFFFFFFFFFFFFF},
                ),
            }
        }

    RETURN_TYPES = ("NOISE",)
    FUNCTION = "get_noise"
    CATEGORY = "sampling/custom_sampling/noise"

    def get_noise(self, noise_seed):
        node_data = create_node_data(
            class_type="RandomNoise",
            inputs={
                "noise_seed": noise_seed,
            },
            outputs={"slot_index": 0},
        )
        model = BizyAirNodeIO(
            self.assigned_id,
            {self.assigned_id: node_data},
        )
        return (model,)


class CLIPSetLastLayer(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "clip": (data_types.CLIP,),
                "stop_at_clip_layer": (
                    "INT",
                    {"default": -1, "min": -24, "max": -1, "step": 1},
                ),
            }
        }

    RETURN_TYPES = (data_types.CLIP,)
    FUNCTION = "set_last_layer"

    CATEGORY = "conditioning"

    def set_last_layer(self, clip: BizyAirNodeIO, stop_at_clip_layer):
        new_clip = clip.copy(new_node_id=self.assigned_id)
        new_clip.add_node_data(
            class_type="CLIPSetLastLayer",
            inputs={
                "stop_at_clip_layer": stop_at_clip_layer,
                "clip": clip,
            },
            outputs={"slot_index": 0},
        )
        return (new_clip,)


class InpaintModelConditioning(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "positive": (data_types.CONDITIONING,),
                "negative": (data_types.CONDITIONING,),
                "vae": (data_types.VAE,),
                "pixels": ("IMAGE",),
                "mask": ("MASK",),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING, data_types.CONDITIONING, "LATENT")
    RETURN_NAMES = ("positive", "negative", "latent")
    # FUNCTION = "encode"

    CATEGORY = "conditioning/inpaint"


#  noise_mask newly added in https://github.com/comfyanonymous/ComfyUI/blob/8f0009aad0591ceee59a147738aa227187b07898/nodes.py#L385
# "noise_mask": ("BOOLEAN", {"default": False, "tooltip": "Add a noise mask to the latent so sampling will only happen within the mask. Might improve results or completely break things depending on the model."}),
class InpaintModelConditioning_v2(InpaintModelConditioning):
    CLASS_TYPE_NAME = "InpaintModelConditioning"

    @classmethod
    def INPUT_TYPES(s):
        ret = super().INPUT_TYPES()
        ret["required"]["noise_mask"] = (
            "BOOLEAN",
            {
                "default": False,
                "tooltip": "Add a noise mask to the latent so sampling will only happen within the mask. Might improve results or completely break things depending on the model.",
            },
        )
        return ret


class SharedLoraLoader(BizyAir_LoraLoader_Legacy):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "share_id": ("STRING", {"default": "share_id"}),
                "lora_name": ([],),
                "model": (data_types.MODEL,),
                "clip": (data_types.CLIP,),
                "strength_model": (
                    "FLOAT",
                    {"default": 1.0, "min": -100.0, "max": 100.0, "step": 0.01},
                ),
                "strength_clip": (
                    "FLOAT",
                    {"default": 1.0, "min": -100.0, "max": 100.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = (data_types.MODEL, data_types.CLIP)
    RETURN_NAMES = ("MODEL", "CLIP")
    FUNCTION = "shared_load_lora"
    CATEGORY = f"{PREFIX}/loaders"
    NODE_DISPLAY_NAME = "Shared Lora Loader"

    @classmethod
    def VALIDATE_INPUTS(cls, share_id: str, lora_name: str):
        if lora_name in folder_paths.filename_path_mapping.get("loras", {}):
            return True

        outs = folder_paths.get_share_filename_list("loras", share_id=share_id)
        if lora_name not in outs:
            raise ValueError(
                f"Lora {lora_name} not found in share {share_id} with {outs}"
            )
        return True

    def shared_load_lora(
        self, model, clip, lora_name, strength_model, strength_clip, **kwargs
    ):
        resolved_path = folder_paths.filename_path_mapping["loras"][lora_name]
        return super().load_lora(
            model=model,
            clip=clip,
            lora_name=resolved_path,
            strength_model=strength_model,
            strength_clip=strength_clip,
        )


class ConditioningCombine(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning_1": (data_types.CONDITIONING,),
                "conditioning_2": (data_types.CONDITIONING,),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    # FUNCTION = "combine"
    NODE_DISPLAY_NAME = "Conditioning (Combine)"
    CATEGORY = "conditioning"


class ConditioningAverage(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning_to": (data_types.CONDITIONING,),
                "conditioning_from": (data_types.CONDITIONING,),
                "conditioning_to_strength": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    # FUNCTION = "addWeighted"
    NODE_DISPLAY_NAME = "Conditioning (Average)"
    CATEGORY = "conditioning"


class ConditioningConcat(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning_to": (data_types.CONDITIONING,),
                "conditioning_from": (data_types.CONDITIONING,),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    # FUNCTION = "concat"
    NODE_DISPLAY_NAME = "Conditioning (Concat)"
    CATEGORY = "conditioning"


class ConditioningSetArea(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning": (data_types.CONDITIONING,),
                "width": (
                    "INT",
                    {"default": 64, "min": 64, "max": MAX_RESOLUTION, "step": 8},
                ),
                "height": (
                    "INT",
                    {"default": 64, "min": 64, "max": MAX_RESOLUTION, "step": 8},
                ),
                "x": (
                    "INT",
                    {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 8},
                ),
                "y": (
                    "INT",
                    {"default": 0, "min": 0, "max": MAX_RESOLUTION, "step": 8},
                ),
                "strength": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 10.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    # FUNCTION = "append"
    NODE_DISPLAY_NAME = "Conditioning (Set Area)"
    CATEGORY = "conditioning"


class ConditioningSetAreaPercentage(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning": (data_types.CONDITIONING,),
                "width": (
                    "FLOAT",
                    {"default": 1.0, "min": 0, "max": 1.0, "step": 0.01},
                ),
                "height": (
                    "FLOAT",
                    {"default": 1.0, "min": 0, "max": 1.0, "step": 0.01},
                ),
                "x": ("FLOAT", {"default": 0, "min": 0, "max": 1.0, "step": 0.01}),
                "y": ("FLOAT", {"default": 0, "min": 0, "max": 1.0, "step": 0.01}),
                "strength": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 10.0, "step": 0.01},
                ),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    # FUNCTION = "append"
    NODE_DISPLAY_NAME = "Conditioning (Set Area with Percentage)"
    CATEGORY = "conditioning"


class ConditioningSetMask(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning": (data_types.CONDITIONING,),
                "mask": ("MASK",),
                "strength": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 10.0, "step": 0.01},
                ),
                "set_cond_area": (["default", "mask bounds"],),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    # FUNCTION = "append"
    NODE_DISPLAY_NAME = "Conditioning (Set Mask)"
    CATEGORY = "conditioning"


class ConditioningZeroOut(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {"required": {"conditioning": (data_types.CONDITIONING,)}}

    RETURN_TYPES = (data_types.CONDITIONING,)
    # FUNCTION = "zero_out"

    CATEGORY = "advanced/conditioning"


class ConditioningSetTimestepRange(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning": (data_types.CONDITIONING,),
                "start": (
                    "FLOAT",
                    {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
                "end": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.001},
                ),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    # FUNCTION = "set_range"

    CATEGORY = "advanced/conditioning"


class SharedControlNetLoader(BizyAir_ControlNetLoader_Legacy):
    @classmethod
    def INPUT_TYPES(s):
        ret = super().INPUT_TYPES()
        ret["required"]["share_id"] = ("STRING", {"default": "share_id"})
        return ret

    NODE_DISPLAY_NAME = "Shared Load ControlNet Model"

    @classmethod
    def VALIDATE_INPUTS(cls, share_id: str, control_net_name: str):
        if control_net_name in folder_paths.filename_path_mapping.get("controlnet", {}):
            return True

        outs = folder_paths.get_share_filename_list("controlnet", share_id=share_id)
        if control_net_name not in outs:
            raise ValueError(
                f"ControlNet {control_net_name} not found in share {share_id} with {outs}"
            )

        return True

    def load_controlnet(self, control_net_name, share_id, **kwargs):
        resolved_path = folder_paths.filename_path_mapping["controlnet"][
            control_net_name
        ]
        return super().load_controlnet(control_net_name=resolved_path, **kwargs)


class CLIPVisionEncode(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {"required": {"clip_vision": ("CLIP_VISION",), "image": ("IMAGE",)}}

    RETURN_TYPES = ("CLIP_VISION_OUTPUT",)
    # FUNCTION = "encode"

    CATEGORY = "conditioning"


class StyleModelLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "style_model_name": (folder_paths.get_filename_list("style_models"),)
                # "style_model_name": (
                #     [
                #         "to choose",
                #     ],
                # ),
                # "model_version_id": (
                #     "STRING",
                #     {
                #         "default": "",
                #     },
                # ),
            }
        }

    RETURN_TYPES = (data_types.STYLE_MODEL,)
    # FUNCTION = "load_style_model"

    CATEGORY = "loaders"


class StyleModelApply(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning": (data_types.CONDITIONING,),
                "style_model": (data_types.STYLE_MODEL,),
                "clip_vision_output": ("CLIP_VISION_OUTPUT",),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    # FUNCTION = "apply_stylemodel"

    CATEGORY = "conditioning/style_model"



import asyncio
import copy
import heapq
import logging
import threading


class UploadQueue:
    def __init__(self):
        self.mutex = threading.RLock()
        self.not_empty = threading.Condition(self.mutex)
        self.task_counter = 0
        self.queue = []
        self.currently_running = {}

    def put(self, item):
        with self.mutex:
            heapq.heappush(self.queue, item)
            self.not_empty.notify()

    def get(self, timeout=None):
        with self.not_empty:
            while len(self.queue) == 0:
                self.not_empty.wait(timeout=timeout)
                if timeout is not None and len(self.queue) == 0:
                    return None
            item = heapq.heappop(self.queue)
            # i = self.task_counter
            upload_id = item["upload_id"]
            self.currently_running[upload_id] = copy.deepcopy(item)
            self.task_counter += 1
            return (item, upload_id)

    def task_done(self, upload_id):
        with self.mutex:
            self.currently_running.pop(upload_id)


def upload_worker(server, q):
    timeout = 1000.0
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    while True:
        try:
            queue_item = q.get(timeout=timeout)
            if queue_item is not None:
                item, upload_id = queue_item
                loop = asyncio.get_event_loop()
                loop.run_until_complete(server.upload_manager.do_upload(item))
                q.task_done(upload_id)
            else:
                continue
        except Exception as e:
            logging.error(f"Failed to upload file: {e}")



import asyncio
import json
import ssl
from typing import Any, Dict, List, Optional


class ConnectionState:
    """连接状态枚举"""

    INIT = "初始化"
    CONNECTING = "连接中"
    CONNECTED = "已连接"
    READING = "读取中"
    CLOSING = "关闭中"
    CLOSED = "已关闭"
    ERROR = "错误"


class StreamResponse:
    """自定义流式响应类，支持状态管理和错误处理"""

    def __init__(self, request_data: Dict[str, Any], timeout: int = 60):
        # 请求和连接数据
        self.request_data = request_data
        self.reader = None
        self.writer = None
        self.timeout = timeout

        # 连接状态管理
        self.state = ConnectionState.INIT
        self.state_lock = asyncio.Lock()  # 状态修改锁
        self.connection_event = asyncio.Event()
        self.error = None

        # 数据相关
        self.content = self
        self.buffer = b""
        self.chunks_received = 0
        self.done_received = False

        # 调试信息
        self.debug_id = id(self)  # 用于日志区分不同实例

    async def _set_state(self, new_state: str, error: Exception = None) -> None:
        """线程安全地设置连接状态"""
        async with self.state_lock:
            self.state = new_state
            if error:
                self.error = error

            # 如果状态为已连接，触发连接事件
            if new_state == ConnectionState.CONNECTED:
                self.connection_event.set()

            # 如果状态为错误，也触发连接事件，让等待连接的代码继续执行并处理错误
            if new_state == ConnectionState.ERROR:
                self.connection_event.set()

    async def connect_and_request(self, api_key: str):
        """建立连接并发送请求，包含严格的状态管理"""
        # 设置状态为连接中
        await self._set_state(ConnectionState.CONNECTING)

        try:
            # 建立SSL连接，添加超时控制
            connect_coro = asyncio.open_connection("api.siliconflow.cn", 443, ssl=True)
            self.reader, self.writer = await asyncio.wait_for(
                connect_coro, timeout=self.timeout
            )

            # 准备HTTP请求
            json_data = json.dumps(self.request_data)
            request = (
                f"POST /v1/chat/completions HTTP/1.1\r\n"
                f"Host: api.siliconflow.cn\r\n"
                f"Authorization: Bearer {api_key}\r\n"
                f"Accept: text/event-stream\r\n"
                f"Content-Type: application/json\r\n"
                f"Content-Length: {len(json_data)}\r\n"
                f"Connection: keep-alive\r\n"
                f"\r\n"
                f"{json_data}"
            )

            # 检查是否仍在连接状态
            async with self.state_lock:
                if self.state != ConnectionState.CONNECTING:
                    raise Exception(f"连接状态已变更为 {self.state}，无法发送请求")

            # 发送请求
            self.writer.write(request.encode("utf-8"))
            await asyncio.wait_for(self.writer.drain(), timeout=self.timeout)

            # 读取HTTP响应头
            response_line = await asyncio.wait_for(
                self.reader.readline(), timeout=self.timeout
            )
            if not response_line:
                raise Exception("服务器关闭了连接")

            status_line = response_line.decode("utf-8").strip()

            if not status_line.startswith("HTTP/1.1 200"):
                raise Exception(f"API请求失败: {status_line}")

            # 读取响应头
            headers = {}
            while True:
                # 再次检查状态
                async with self.state_lock:
                    if self.state != ConnectionState.CONNECTING:
                        raise Exception(
                            f"连接状态已变更为 {self.state}，停止读取响应头"
                        )

                line = await asyncio.wait_for(
                    self.reader.readline(), timeout=self.timeout
                )
                if line == b"\r\n" or not line:
                    break

                try:
                    header_line = line.decode("utf-8").strip()
                    if ":" in header_line:
                        key, value = header_line.split(":", 1)
                        headers[key.strip()] = value.strip()
                except Exception:
                    pass

            # 成功连接
            await self._set_state(ConnectionState.CONNECTED)

        except asyncio.TimeoutError as e:
            await self._set_state(ConnectionState.ERROR, e)
            await self.release()
            raise Exception(f"连接超时 (>{self.timeout}秒)") from e

        except Exception as e:
            await self._set_state(ConnectionState.ERROR, e)
            await self.release()
            raise Exception(f"连接失败: {str(e)}") from e

    async def iter_any(self):
        """模拟aiohttp的iter_any方法"""
        # 等待连接完成
        if not self.connection_event.is_set():
            try:
                await asyncio.wait_for(
                    self.connection_event.wait(), timeout=self.timeout
                )
            except asyncio.TimeoutError as e:
                await self._set_state(ConnectionState.ERROR, e)
                await self.release()
                raise Exception(f"等待连接超时 (>{self.timeout}秒)") from e

        # 检查连接状态
        async with self.state_lock:
            if self.state == ConnectionState.ERROR:
                error_msg = f"连接发生错误: {str(self.error)}"
                raise Exception(error_msg) from self.error

            if self.state != ConnectionState.CONNECTED:
                error_msg = f"连接状态为 {self.state}，无法读取数据"
                raise Exception(error_msg)

            # 设置状态为读取中
            self.state = ConnectionState.READING

        # 读取数据
        try:
            max_empty_chunks = 3  # 连续空块计数，防止无限循环
            empty_chunk_count = 0

            while True:
                # 检查状态
                async with self.state_lock:
                    if self.state != ConnectionState.READING:
                        break

                try:
                    # 读取一段数据，添加超时控制
                    chunk = await asyncio.wait_for(
                        self.reader.read(1024), timeout=self.timeout
                    )

                    # 检测连续空块，防止无限循环
                    if not chunk:
                        empty_chunk_count += 1
                        if empty_chunk_count >= max_empty_chunks:
                            break
                        # 给个短暂的休息，避免CPU过度使用
                        await asyncio.sleep(0.05)
                        continue
                    else:
                        empty_chunk_count = 0

                    # 更新计数并生成数据
                    self.chunks_received += 1
                    yield chunk

                    # 检查是否接收到了结束标记 [DONE]
                    if b"data: [DONE]" in chunk:
                        self.done_received = True
                        break

                except asyncio.TimeoutError as e:
                    await self._set_state(ConnectionState.ERROR, e)
                    break

                except (ConnectionError, OSError) as e:
                    await self._set_state(ConnectionState.ERROR, e)
                    break

        except Exception as e:
            await self._set_state(ConnectionState.ERROR, e)
            error_msg = f"读取数据失败: {str(e)}"
            raise
        finally:
            # 关闭连接
            await self.release()

    async def release(self):
        """安全关闭连接并释放资源"""
        async with self.state_lock:
            # 检查是否已经在关闭或已关闭状态
            if self.state in [ConnectionState.CLOSING, ConnectionState.CLOSED]:
                return

            # 设置状态为关闭中
            self.state = ConnectionState.CLOSING

        # 关闭写入端
        if self.writer:
            writer = self.writer
            self.writer = None  # 清除引用
            self.reader = None  # 清除引用

            try:
                # 尝试优雅关闭
                if not writer.is_closing():
                    # 先确保发送所有数据
                    try:
                        await asyncio.wait_for(writer.drain(), timeout=1.0)
                    except (asyncio.TimeoutError, ConnectionError, OSError):
                        # 忽略drain错误
                        pass

                    # 写入EOF标记
                    try:
                        writer.write_eof()
                    except (OSError, RuntimeError):
                        # 忽略无法写入EOF的错误
                        pass

                    # 关闭连接
                    writer.close()

                # 等待连接关闭完成
                try:
                    await asyncio.wait_for(writer.wait_closed(), timeout=2.0)
                except (asyncio.TimeoutError, ConnectionError, OSError, RuntimeError):
                    # 忽略等待关闭的错误
                    pass
            except Exception:
                pass

        # 设置为已关闭状态
        await self._set_state(ConnectionState.CLOSED)



import configparser
import os
import shutil

from .errno import errnos


class UserProfile:
    def __init__(self):
        UserProfile.instance = self

        self.profile_cache = {}
        self.load_profile()
        self.lang = "zh"

    def getLang(self):
        return self.lang

    def getAll(self):
        return self.profile_cache

    def load_profile(self):
        """加载用户配置文件到缓存"""
        profile_path = os.path.join(os.getenv("BIZYAIR_COMFYUI_PATH"), "profile.ini")
        example_path = os.path.join(
            os.path.dirname(profile_path), "profile.ini.example"
        )

        # 如果配置文件不存在且示例文件存在，则复制示例文件
        if not os.path.exists(profile_path) and os.path.exists(example_path):
            try:
                shutil.copy2(example_path, profile_path)
            except Exception as e:
                print(
                    f"\033[31m[BizyAir]\033[0m Fail to copy example profile: {str(e)}"
                )
                return

        # 如果配置文件仍不存在，则创建默认配置
        if not os.path.exists(profile_path):
            try:
                config = configparser.ConfigParser()
                config.add_section("global")
                config.set("global", "lang", "zh")
                with open(profile_path, "w") as f:
                    config.write(f)
            except Exception as e:
                print(
                    f"\033[31m[BizyAir]\033[0m Fail to create default profile: {str(e)}"
                )
                return

        try:
            config = configparser.ConfigParser()
            config.read(profile_path)
            self.profile_cache = {}
            for section in config.sections():
                self.profile_cache[section] = {}
                for key, value in config.items(section):
                    self.profile_cache[section][key] = value

            if self.profile_cache["global"]["lang"]:
                self.lang = self.profile_cache["global"]["lang"]
        except Exception as e:
            print(f"\033[31m[BizyAir]\033[0m Fail to read profile: {str(e)}")

    def update_profile(self, json_data):
        profile_path = os.path.join(os.getenv("BIZYAIR_COMFYUI_PATH"), "profile.ini")
        try:
            config = configparser.ConfigParser()
            if os.path.exists(profile_path):
                config.read(profile_path)

            for section, values in json_data.items():
                if not config.has_section(section):
                    config.add_section(section)
                for key, value in values.items():
                    config.set(section, key, str(value))

            with open(profile_path, "w") as f:
                config.write(f)

            # 重新加载配置
            self.load_profile()
        except Exception as e:
            print(f"\033[31m[BizyAir]\033[0m Fail to write profile: {str(e)}")
            return errnos.WRITE_PROFILE_FAILED


user_profile = UserProfile()



import copy
import json
import os
import pprint
import re
import warnings
from collections import defaultdict
from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Collection, Dict, List, Union

from bizyengine.core.common import client, fetch_models_by_type
from bizyengine.core.common.env_var import BIZYAIR_DEBUG, BIZYAIR_SERVER_ADDRESS
from bizyengine.core.configs.conf import ModelRule, config_manager
from bizyengine.core.path_utils.utils import (
    filter_files_extensions,
    get_service_route,
    load_yaml_config,
)

supported_pt_extensions: set[str] = {
    ".ckpt",
    ".pt",
    ".bin",
    ".pth",
    ".safetensors",
    ".pkl",
    ".sft",
}
ScanPathType = list[str]
folder_names_and_paths: dict[str, ScanPathType] = defaultdict(list)
filename_path_mapping: dict[str, dict[str, str]] = {}


@dataclass
class RefreshSettings:
    loras: bool = True
    controlnet: bool = True

    def get(self, folder_name: str, default: bool = True):
        return getattr(self, folder_name, default)

    def set(self, folder_name: str, value: bool):
        setattr(self, folder_name, value)


refresh_settings = RefreshSettings()


def enable_refresh_options(folder_names: Union[str, list[str]]):
    if isinstance(folder_names, str):
        folder_names = [folder_names]
    for folder_name in folder_names:
        refresh_settings.set(folder_name, True)


def disable_refresh_options(folder_names: Union[str, list[str]]):
    if isinstance(folder_names, str):
        folder_names = [folder_names]
    for folder_name in folder_names:
        refresh_settings.set(folder_name, False)


def _get_config_path():
    src_bizyair_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    configs_path = os.path.join(src_bizyair_path, "configs")
    return configs_path


configs_path = _get_config_path()

models_config: Dict[str, Dict[str, Any]] = load_yaml_config(
    os.path.join(configs_path, "models.yaml")
)


def detect_model(model_version_id, detection_type, **kwargs):
    return "SDXL"
    # TODO support detect_model_type
    # json_data = {
    #     "prompt": {
    #         "model_version_id": model_version_id,
    #         "detection_type": detection_type,
    #     },
    #     "exec_info": {"api_key": client.get_api_key()},
    # }
    # detect_model_type: dict = models_config["model_version_config"]["detect_model_type"]
    # resq = client.send_request(
    #     url=detect_model_type["url"], data=json.dumps(json_data).encode("utf-8")
    # )
    # if resq["type"] != "success":
    #     raise RuntimeError(
    #         f"Request failed: {resq.get('type', 'unknown error')} - {resq.get('message', 'No details available')}"
    #     )

    # payload = resq["data"]["payload"]

    # model_type = payload.get("model_type")
    # return model_type


def action_call(action: str, *args, **kwargs) -> any:
    if action == "detect_model":
        return detect_model(*args, **kwargs)


def guess_url_from_node(
    node: Dict[str, Dict[str, Any]], class_type_table: Dict[str, bool]
) -> Union[List[ModelRule], None]:
    rules: List[ModelRule] = config_manager.get_rules(node["class_type"])
    out = []
    model_version_id_prefix = models_config["model_version_config"][
        "model_version_id_prefix"
    ]
    model_type = None
    for rule in rules:
        if len(rule.inputs) == 0:
            out.append(rule)

        skip = False
        for key, patterns in rule.inputs.items():
            if skip:
                break
            for pattern in patterns:
                value = node["inputs"][key]
                if isinstance(pattern, str) and re.search(pattern, value) is not None:
                    out.append(rule)
                    skip = True
                    break
                elif isinstance(pattern, dict) and value.startswith(
                    model_version_id_prefix
                ):
                    action = pattern["action"]
                    detection_type = pattern["detection_type"]
                    model_type = action_call(
                        action=action,
                        model_version_id=value[len(model_version_id_prefix) :],
                        detection_type=detection_type,
                    )
        if model_type and model_type == rule.base_model:
            out.append(rule)

    return out


def guess_config(
    *,
    ckpt_name: str = None,
    unet_name: str = None,
    vae_name: str = None,
    clip_name: str = None,
) -> str:
    warnings.warn("The interface has changed, please do not use it", DeprecationWarning)


def get_config_file_list(base_path=None) -> list:
    if base_path is None:
        base_path = os.path.dirname(os.path.abspath(__file__))
    path = os.path.join(configs_path)
    extensions = ".yaml"
    config_files = []
    for root, _, files in os.walk(path):
        for file in files:
            if file.endswith(extensions):
                file_path = os.path.join(root, file)
                config_files.append(file_path)
    return config_files


def cached_filename_list(
    folder_name: str, *, share_id: str = None, verbose=False, refresh=False
) -> list[str]:
    global filename_path_mapping
    if refresh or folder_name not in filename_path_mapping:
        model_types: Dict[str, str] = models_config["model_types"]
        if share_id:
            url = f"{BIZYAIR_SERVER_ADDRESS}/{share_id}/models/files"
        else:
            url = get_service_route(models_config["model_hub"]["find_model"])
        msg = fetch_models_by_type(
            url=url, method="GET", model_type=model_types[folder_name]
        )
        if verbose:
            pprint.pprint({"cached_filename_list": msg})

        try:
            if not msg or "data" not in msg or msg["data"] is None:
                return []

            filename_path_mapping[folder_name] = {
                x["label_path"]: x["real_path"]
                for x in msg["data"]["files"]
                if x["label_path"]
            }
        except Exception as e:
            warnings.warn(f"Failed to get filename list: {e}")
            return []
        finally:
            # TODO fix share_id vaild refresh settings
            if share_id is None:
                disable_refresh_options(folder_name)

    return list(
        filter_files_extensions(
            filename_path_mapping[folder_name].keys(),
            extensions=supported_pt_extensions,
        )
    )


def convert_prompt_label_path_to_real_path(prompt: dict[str, dict[str, any]]) -> dict:
    # TODO fix Temporarily write dead
    new_prompt = {}
    for unique_id in prompt:
        new_prompt[unique_id] = copy.copy(prompt[unique_id])
        inputs = copy.copy(prompt[unique_id]["inputs"])

        for key, folder_name in [
            ("lora_name", "loras"),
            ("control_net_name", "controlnet"),
        ]:
            if key in inputs:
                value = inputs[key]
                new_value = filename_path_mapping.get(folder_name, {}).get(value, None)
                if new_value:
                    inputs[key] = new_value
                else:
                    file_list = get_filename_list(folder_name)
                    if value not in file_list:
                        raise ValueError(
                            f"{key} '{value}' not found in file list. Available {key} names: {', '.join(file_list)}"
                        )

        new_prompt[unique_id]["inputs"] = inputs
    return new_prompt


def get_share_filename_list(folder_name, share_id, *, verbose=BIZYAIR_DEBUG):
    assert share_id is not None and isinstance(share_id, str)
    # TODO fix share_id vaild refresh settings
    return cached_filename_list(
        folder_name, share_id=share_id, verbose=verbose, refresh=True
    )


def get_filename_list(folder_name, *, verbose=BIZYAIR_DEBUG):

    global folder_names_and_paths
    results = folder_names_and_paths.get(folder_name, [])
    # 社区node上线后移除
    # if folder_name in models_config["model_types"]:
    #     refresh = refresh_settings.get(folder_name, True)
    #     results.extend(
    #         cached_filename_list(folder_name, verbose=verbose, refresh=refresh)
    #     )
    # if folder_name in folder_names_and_paths:
    #     results.extend(folder_names_and_paths[folder_name])
    # if BIZYAIR_DEBUG:
    #     try:
    #         import folder_paths
    #
    #         results.extend(folder_paths.get_filename_list(folder_name))
    #     except:
    #         pass
    return results


def filter_files_extensions(
    files: Collection[str], extensions: Collection[str]
) -> list[str]:
    return sorted(
        list(
            filter(
                lambda a: os.path.splitext(a)[-1].lower() in extensions
                or len(extensions) == 0,
                files,
            )
        )
    )


def recursive_extract_models(data: Any, prefix_path: str = "") -> List[str]:
    def merge_paths(base_path: str, new_path: Any) -> str:
        if not isinstance(new_path, str):
            return base_path
        return f"{base_path}/{new_path}" if base_path else new_path

    results: List[str] = []
    if isinstance(data, dict):
        for key, value in data.items():
            new_prefix = merge_paths(prefix_path, key)
            results.extend(recursive_extract_models(value, new_prefix))
    elif isinstance(data, list):
        for item in data:
            new_prefix = merge_paths(prefix_path, item)
            results.extend(recursive_extract_models(item, new_prefix))
    elif isinstance(data, str) and prefix_path.endswith(data):
        return filter_files_extensions([prefix_path], supported_pt_extensions)

    return results


def load_json(file_path: str) -> dict:
    with open(file_path, "r") as file:
        data = json.load(file)
    return data


def init_config():
    global folder_names_and_paths
    for k, filenames in config_manager.model_path_manager.model_paths.items():
        folder_names_and_paths[k].extend(filenames)
    if BIZYAIR_DEBUG:
        pprint.pprint("=" * 20 + "init_config: " + "=" * 20)
        pprint.pprint(folder_names_and_paths)


init_config()


if __name__ == "__main__":
    # print(f"Loaded config from {get_config_file_list()}")
    # configs = [load_yaml_config(x) for x in get_config_file_list()]
    # print(get_filename_list("clip_vision"))
    # print(folder_names_and_paths)

    api_key = os.getenv("BIZYAIR_API_KEY", "")
    host_ckpts = get_filename_list("loras", verbose=True)
    print(host_ckpts)



import bizyengine.core.path_utils as path_utils
from bizyengine.core.common import set_api_key, validate_api_key
from bizyengine.core.nodes_base import (
    NODE_CLASS_MAPPINGS,
    NODE_DISPLAY_NAME_MAPPINGS,
    BizyAirBaseNode,
)
from bizyengine.core.nodes_io import BizyAirNodeIO, create_node_data



import os
import uuid

import folder_paths
import requests
import torch
from bizyengine.core import BizyAirBaseNode


class BizyAir_IF_TrellisCheckpointLoader(BizyAirBaseNode):
    """
    Node to manage the loading of the TRELLIS model.
    Follows ComfyUI conventions for model management.
    """

    @classmethod
    def INPUT_TYPES(cls):
        """Define input types with device-specific options."""
        device_options = ["cuda"]

        return {
            "required": {
                "model_name": (["TRELLIS-image-large"],),
                "dinov2_model": (
                    ["dinov2_vitl14_reg", "dinov2_vitg14_reg"],
                    {
                        "default": "dinov2_vitl14_reg",
                        "tooltip": "Select the Dinov2 model to use for the image to 3D conversion. Smaller models work but better results with larger models.",
                    },
                ),
                "use_fp16": ("BOOLEAN", {"default": True}),
                "attn_backend": (
                    ["sage", "xformers", "flash_attn", "sdpa", "naive"],
                    {
                        "default": "sage",
                        "tooltip": "Select the attention backend to use for the image to 3D conversion. Sage is experimental but faster",
                    },
                ),
                "smooth_k": (
                    "BOOLEAN",
                    {
                        "default": True,
                        "tooltip": "Smooth k for sage attention. This is a hyperparameter that controls the smoothness of the attention distribution. It is a boolean value that determines whether to use smooth k or not. Smooth k is a hyperparameter that controls the smoothness of the attention distribution. It is a boolean value that determines whether to use smooth k or not.",
                    },
                ),
                "spconv_algo": (
                    ["implicit_gemm", "native"],
                    {
                        "default": "implicit_gemm",
                        "tooltip": "Select the spconv algorithm to use for the image to 3D conversion. Implicit gemm is the best but slower. Native is the fastest but less accurate.",
                    },
                ),
                "main_device": (device_options, {"default": device_options[0]}),
            },
        }

    RETURN_TYPES = ("TRELLIS_MODEL",)
    RETURN_NAMES = ("model",)
    CATEGORY = "☁️BizyAir/Trellis"
    NODE_DISPLAY_NAME = "☁️BizyAir Load Trellis"


class BizyAir_IF_TrellisImageTo3D(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": ("TRELLIS_MODEL",),
                "mode": (
                    ["single", "multi"],
                    {
                        "default": "single",
                        "tooltip": "Mode. single is a single image. with multi you can provide multiple reference angles for the 3D model",
                    },
                ),
                "images": ("IMAGE", {"list": True}),
                "seed": ("INT", {"default": 0, "min": 0, "max": 0x7FFFFFFF}),
                "ss_guidance_strength": (
                    "FLOAT",
                    {"default": 7.5, "min": 0.0, "max": 12.0, "step": 0.1},
                ),
                "ss_sampling_steps": ("INT", {"default": 12, "min": 1, "max": 50}),
                "slat_guidance_strength": (
                    "FLOAT",
                    {"default": 3.0, "min": 0.0, "max": 12.0, "step": 0.1},
                ),
                "slat_sampling_steps": ("INT", {"default": 12, "min": 1, "max": 50}),
                "multimode": (
                    ["stochastic", "multidiffusion"],
                    {"default": "stochastic"},
                ),
            },
            "optional": {
                "masks": ("MASK", {"list": True}),
            },
        }

    RETURN_TYPES = (
        "trellis_gaussian",
        "trellis_mesh",
    )
    CATEGORY = "☁️BizyAir/Trellis"
    NODE_DISPLAY_NAME = "☁️BizyAir Trellis Predict"
    OUTPUT_NODE = True


class BizyAir_Trans3D2GlbFile(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "trellis_gaussian": ("trellis_gaussian", {"forceInput": True}),
                "trellis_mesh": ("trellis_mesh", {"forceInput": True}),
                "mesh_simplify": (
                    "FLOAT",
                    {
                        "default": 0.95,
                        "min": 0.9,
                        "max": 1.0,
                        "step": 0.01,
                        "tooltip": "Simplify the mesh. the lower the value more polygons the mesh will have",
                    },
                ),
                "texture_size": (
                    "INT",
                    {
                        "default": 1024,
                        "min": 512,
                        "max": 2048,
                        "step": 512,
                        "tooltip": "Texture size. the higher the value the more detailed the texture will be",
                    },
                ),
                "texture_mode": (
                    ["blank", "fast", "opt"],
                    {
                        "default": "fast",
                        "tooltip": "Texture mode. blank is no texture. fast is a fast texture. opt is a high quality texture",
                    },
                ),
                "save_glb": (
                    "BOOLEAN",
                    {
                        "default": True,
                        "tooltip": "Save the GLB file this is the 3D model",
                    },
                ),
                "save_texture": (
                    "BOOLEAN",
                    {"default": False, "tooltip": "Save the texture file"},
                ),
            }
        }

    CATEGORY = "☁️BizyAir/Trellis"
    NODE_DISPLAY_NAME = "☁️BizyAir Generate Glb With Trellis"
    RETURN_TYPES = ("STRING", "IMAGE")
    RETURN_NAMES = ("url", "texture_image")


class BizyAirDownloadFile(BizyAirBaseNode):
    NODE_DISPLAY_NAME = "☁️BizyAir Download File"

    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "url": ("STRING", {"default": ""}),
                "file_name": ("STRING", {"default": "default"}),
            }
        }

    CATEGORY = "☁️BizyAir/Trellis"
    FUNCTION = "main"

    RETURN_TYPES = ("STRING",)
    RETURN_NAMES = ("path",)
    OUTPUT_NODE = True
    OUTPUT_IS_LIST = (False,)

    def main(self, url, file_name):
        assert url is not None
        file_name = file_name + ".glb"
        out_dir = os.path.join(folder_paths.get_output_directory(), "trellis_output")
        os.makedirs(out_dir, exist_ok=True)
        local_path = os.path.join(out_dir, file_name)
        output = os.path.join("trellis_output", file_name)
        response = requests.get(url)
        if response.status_code == 200:
            with open(local_path, "wb") as file:
                file.write(response.content)
            print("download finished in {}".format(local_path))
        else:
            print(f"download error: {response.status_code}")
        return (output,)

    @classmethod
    def IS_CHANGED(self, url, file_name, *args, **kwargs):
        return uuid.uuid4().hex



from bizyengine.core import BizyAirBaseNode, BizyAirNodeIO, data_types
from bizyengine.core.configs.conf import config_manager
from bizyengine.core.path_utils import path_manager as folder_paths


class NunchakuFluxDiTLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        attention_options = ["nunchaku-fp16"]
        dtype_options = ["bfloat16"]
        return {
            "required": {
                "model_path": (
                    ["svdq-int4-flux.1-dev"],
                    {"tooltip": "The SVDQuant quantized FLUX.1 models."},
                ),
                "cache_threshold": (
                    "FLOAT",
                    {
                        "default": 0,
                        "min": 0,
                        "max": 1,
                        "step": 0.001,
                        "tooltip": "Adjusts the caching tolerance like `residual_diff_threshold` in WaveSpeed. "
                        "Increasing the value enhances speed at the cost of quality. "
                        "A typical setting is 0.12. Setting it to 0 disables the effect.",
                    },
                ),
                "attention": (
                    attention_options,
                    {
                        "default": attention_options[0],
                        "tooltip": "Attention implementation. The default implementation is `flash-attention2`. "
                        "`nunchaku-fp16` use FP16 attention, offering ~1.2× speedup. "
                        "Note that 20-series GPUs can only use `nunchaku-fp16`.",
                    },
                ),
                "cpu_offload": (
                    ["disable"],
                    {
                        "default": "auto",
                        "tooltip": "Whether to enable CPU offload for the transformer model."
                        "auto' will enable it if the GPU memory is less than 14G.",
                    },
                ),
                "device_id": (
                    "INT",
                    {
                        "default": 0,
                        "min": 0,
                        "max": 0,
                        "step": 1,
                        "display": "number",
                        "lazy": True,
                        "tooltip": "The GPU device ID to use for the model.",
                    },
                ),
                "data_type": (
                    dtype_options,
                    {
                        "default": dtype_options[0],
                        "tooltip": "Specifies the model's data type. Default is `bfloat16`. "
                        "For 20-series GPUs, which do not support `bfloat16`, use `float16` instead.",
                    },
                ),
            },
            "optional": {
                "i2f_mode": (
                    ["enabled", "always"],
                    {
                        "default": "enabled",
                        "tooltip": "The GEMM implementation for 20-series GPUs"
                        "— this option is only applicable to these GPUs.",
                    },
                )
            },
        }

    RETURN_TYPES = (data_types.MODEL,)
    CATEGORY = "Nunchaku"
    NODE_DISPLAY_NAME = "Nunchaku FLUX DiT Loader"


class NunchakuTextEncoderLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model_type": (["flux"],),
                "text_encoder1": (folder_paths.get_filename_list("text_encoders"),),
                "text_encoder2": (folder_paths.get_filename_list("text_encoders"),),
                "t5_min_length": (
                    "INT",
                    {
                        "default": 512,
                        "min": 256,
                        "max": 1024,
                        "step": 128,
                        "display": "number",
                        "lazy": True,
                    },
                ),
                "use_4bit_t5": (["disable"],),
                "int4_model": (
                    ["none"],
                    {"tooltip": "The name of the 4-bit T5 model."},
                ),
            }
        }

    RETURN_TYPES = (data_types.CLIP,)
    # FUNCTION = "load_text_encoder"
    CATEGORY = "Nunchaku"
    NODE_DISPLAY_NAME = "Nunchaku Text Encoder Loader"


class NunchakuFluxLoraLoader(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": (
                    data_types.MODEL,
                    {"tooltip": "The diffusion model the LoRA will be applied to."},
                ),
                "lora_name": (
                    [
                        "to choose",
                    ],
                    {"tooltip": "The name of the LoRA."},
                ),
                "lora_strength": (
                    "FLOAT",
                    {
                        "default": 1.0,
                        "min": -100.0,
                        "max": 100.0,
                        "step": 0.01,
                        "tooltip": "How strongly to modify the diffusion model. This value can be negative.",
                    },
                ),
                "model_version_id": (
                    "STRING",
                    {
                        "default": "",
                    },
                ),
            }
        }

    RETURN_TYPES = (data_types.MODEL,)
    OUTPUT_TOOLTIPS = ("The modified diffusion model.",)
    FUNCTION = "load_lora"
    NODE_DISPLAY_NAME = "Nunchaku FLUX.1 LoRA Loader"

    CATEGORY = "Nunchaku"
    DESCRIPTION = (
        "LoRAs are used to modify the diffusion model, "
        "altering the way in which latents are denoised such as applying styles. "
        "You can link multiple LoRA nodes."
    )

    @classmethod
    def VALIDATE_INPUTS(cls, lora_name, **kwargs):
        if lora_name == "" or lora_name is None:
            return False
        return True

    def load_lora(
        self,
        model,
        lora_name,
        lora_strength,
        model_version_id: str = None,
    ):
        assigned_id = self.assigned_id
        new_model: BizyAirNodeIO = model.copy(assigned_id)

        if model_version_id is not None and model_version_id != "":
            # use model version id as lora name
            lora_name = (
                f"{config_manager.get_model_version_id_prefix()}{model_version_id}"
            )
        new_model.add_node_data(
            class_type="NunchakuFluxLoraLoader",
            inputs={
                "model": model,
                "lora_name": lora_name,
                "lora_strength": lora_strength,
            },
            outputs={"slot_index": 0},
        )
        return (new_model,)



"""
Reference: https://github.com/Jonseed/ComfyUI-Detail-Daemon
"""

from bizyengine.core import BizyAirBaseNode


class DetailDaemonSamplerNode(BizyAirBaseNode):
    DESCRIPTION = "This sampler wrapper works by adjusting the sigma passed to the model, while the rest of sampling stays the same."
    CATEGORY = "sampling/custom_sampling/samplers"
    RETURN_TYPES = ("SAMPLER",)
    # FUNCTION = "go"

    NODE_DISPLAY_NAME = "Detail Daemon Sampler"

    @classmethod
    def INPUT_TYPES(cls) -> dict:
        return {
            "required": {
                "sampler": ("SAMPLER",),
                "detail_amount": (
                    "FLOAT",
                    {"default": 0.1, "min": -5.0, "max": 5.0, "step": 0.01},
                ),
                "start": (
                    "FLOAT",
                    {"default": 0.2, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
                "end": (
                    "FLOAT",
                    {"default": 0.8, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
                "bias": (
                    "FLOAT",
                    {"default": 0.5, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
                "exponent": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 10.0, "step": 0.05},
                ),
                "start_offset": (
                    "FLOAT",
                    {"default": 0.0, "min": -1.0, "max": 1.0, "step": 0.01},
                ),
                "end_offset": (
                    "FLOAT",
                    {"default": 0.0, "min": -1.0, "max": 1.0, "step": 0.01},
                ),
                "fade": (
                    "FLOAT",
                    {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.05},
                ),
                "smooth": ("BOOLEAN", {"default": True}),
                "cfg_scale_override": (
                    "FLOAT",
                    {
                        "default": 0,
                        "min": 0.0,
                        "max": 100.0,
                        "step": 0.5,
                        "round": 0.01,
                        "tooltip": "If set to 0, the sampler will automatically determine the CFG scale (if possible). Set to some other value to override.",
                    },
                ),
            },
        }


class MultiplySigmas(BizyAirBaseNode):
    # FUNCTION = "simple_output"
    RETURN_TYPES = ("SIGMAS",)
    CATEGORY = "sampling/custom_sampling/sigmas"
    NODE_DISPLAY_NAME = "Multiply Sigmas (stateless)"

    @classmethod
    def INPUT_TYPES(cls) -> dict:
        return {
            "required": {
                "sigmas": ("SIGMAS", {"forceInput": True}),
                "factor": (
                    "FLOAT",
                    {"default": 1, "min": 0, "max": 100, "step": 0.001},
                ),
                "start": ("FLOAT", {"default": 0, "min": 0, "max": 1, "step": 0.001}),
                "end": ("FLOAT", {"default": 1, "min": 0, "max": 1, "step": 0.001}),
            }
        }


class LyingSigmaSampler(BizyAirBaseNode):
    CATEGORY = "sampling/custom_sampling"
    RETURN_TYPES = ("SAMPLER",)
    NODE_DISPLAY_NAME = "Lying Sigma Sampler"
    # FUNCTION = "go"

    @classmethod
    def INPUT_TYPES(cls) -> dict:
        return {
            "required": {
                "sampler": ("SAMPLER",),
                "dishonesty_factor": (
                    "FLOAT",
                    {
                        "default": -0.05,
                        "min": -0.999,
                        "step": 0.01,
                        "tooltip": "Multiplier for sigmas passed to the model. -0.05 means we reduce the sigma by 5%.",
                    },
                ),
            },
            "optional": {
                "start_percent": (
                    "FLOAT",
                    {"default": 0.1, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
                "end_percent": (
                    "FLOAT",
                    {"default": 0.9, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
            },
        }


class DetailDaemonGraphSigmasNode(BizyAirBaseNode):
    RETURN_TYPES = ()
    OUTPUT_NODE = True
    CATEGORY = "sampling/custom_sampling/sigmas"
    NODE_DISPLAY_NAME = "Detail Daemon Graph Sigmas"
    # FUNCTION = "make_graph"

    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "sigmas": ("SIGMAS", {"forceInput": True}),
                "detail_amount": (
                    "FLOAT",
                    {"default": 0.1, "min": -5.0, "max": 5.0, "step": 0.01},
                ),
                "start": (
                    "FLOAT",
                    {"default": 0.2, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
                "end": (
                    "FLOAT",
                    {"default": 0.8, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
                "bias": (
                    "FLOAT",
                    {"default": 0.5, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
                "exponent": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 10.0, "step": 0.05},
                ),
                "start_offset": (
                    "FLOAT",
                    {"default": 0.0, "min": -1.0, "max": 1.0, "step": 0.01},
                ),
                "end_offset": (
                    "FLOAT",
                    {"default": 0.0, "min": -1.0, "max": 1.0, "step": 0.01},
                ),
                "fade": (
                    "FLOAT",
                    {"default": 0.0, "min": 0.0, "max": 1.0, "step": 0.05},
                ),
                "smooth": ("BOOLEAN", {"default": True}),
                "cfg_scale": (
                    "FLOAT",
                    {
                        "default": 1.0,
                        "min": 0.0,
                        "max": 100.0,
                        "step": 0.5,
                        "round": 0.01,
                    },
                ),
            },
        }



import json
import time
import warnings

import requests

try:
    from comfy_api_nodes.apinode_utils import (
        download_url_to_video_output,
        tensor_to_base64_string,
    )
except ModuleNotFoundError as e:
    download_url_to_video_output = None
    tensor_to_base64_string = None

    ERROR_MSG = f"Error {e} ComfyUI API nodes module not found. Please ensure you have ComfyUI version 0.3.36 or later installed."

    warnings.warn(ERROR_MSG)


from ..core.common import client
from ..core.common.env_var import BIZYAIR_DEBUG
from ..core.nodes_base import BizyAirBaseNode


class WanApiNodeBase:
    MODEL_ENDPOINTS = {
        "Wan-AI/Wan2.1-I2V-14B-480P-Diffusers": "https://bizyair-api.siliconflow.cn/x/v1/supernode/faas-wan-i2v-14b-480p-server"
    }


class Wan_LoraLoader(BizyAirBaseNode):

    @classmethod
    def INPUT_TYPES(cls):
        return {
            "required": {
                "lora_name": (
                    "STRING",
                    {
                        "multiline": True,
                        "default": "https://huggingface.co/Remade-AI/Squish/resolve/main/squish_18.safetensors",
                        "tooltip": "LoRA 模型下载地址",
                    },
                ),
            },
            "optional": {
                "lora_weight": (
                    "FLOAT",
                    {
                        "default": 1,
                        "min": 0.0,
                        "max": 1.0,
                        "step": 0.05,
                        "tooltip": "LoRA权重强度",
                    },
                )
            },
        }

    RETURN_TYPES = ("LORA_CONFIG",)
    RETURN_NAMES = ("lora_config",)
    FUNCTION = "apply_lora"
    CATEGORY = "Diffusers/WAN Video Generation"

    def apply_lora(self, lora_name, lora_weight=0.75):
        return ([(lora_name, lora_weight)],)


class Wan_ImageToVideoPipeline(WanApiNodeBase, BizyAirBaseNode):

    POLLING_INTERVAL = 10  # sec
    MAX_POLLING_TIME = 60 * 20  # sec

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "image": (
                    "IMAGE",
                    {
                        "default": None,
                        "tooltip": "Optional reference image to guide video generation",
                    },
                ),
                "model_id": (["Wan-AI/Wan2.1-I2V-14B-480P-Diffusers"],),
                "prompt": (
                    "STRING",
                    {
                        "multiline": True,
                        "default": "",
                        "tooltip": "Text description of the video",
                    },
                ),
            },
            "optional": {
                "negative_prompt": (
                    "STRING",
                    {
                        "multiline": True,
                        "default": "",
                        "tooltip": "Negative text prompt to guide what to avoid in the video",
                    },
                ),
                "steps": ("INT", {"default": 30, "min": 1, "max": 40}),
                "cfg": (
                    "FLOAT",
                    {
                        "default": 6.0,
                        "min": 0.0,
                        "max": 100.0,
                        "step": 0.1,
                        "round": 0.01,
                    },
                ),
                "seed": (
                    "INT",
                    {
                        "default": 0,
                        "min": 0,
                        "max": 0xFFFFFFFF,
                        "step": 1,
                        "display": "number",
                        "control_after_generate": True,
                        "tooltip": "Seed for video generation (0 for random)",
                    },
                ),
                "use_teacache": (["enable", "disable"],),
                "lora_config": ("LORA_CONFIG", {}),
            },
        }

    RETURN_TYPES = ("VIDEO",)

    FUNCTION = "generate_video"
    CATEGORY = "Diffusers/WAN Video Generation"

    def _encode_image(self, image_tensor):
        # https://docs.comfy.org/custom-nodes/backend/snippets
        tensor_to_base64_string(image_tensor=image_tensor, mime_type="image/webp")
        base64_str = "data:image/webp;base64," + tensor_to_base64_string(
            image_tensor=image_tensor, mime_type="image/webp"
        )
        return base64_str

    def _prepare_headers(self):
        headers = client._headers()
        headers["X-Fn-Task-Mode"] = "non-blocking"
        return headers

    def _send_initial_request(self, endpoint, request_data):
        headers = self._prepare_headers()
        response = client.send_request(
            url=endpoint,
            data=json.dumps({"prompt": request_data}).encode(),
            headers=headers,
        )
        return response["query_url"]

    def _poll_for_completion(self, query_url):
        start_time = time.time()
        headers = self._prepare_headers()

        while time.time() - start_time < self.MAX_POLLING_TIME:
            response = requests.get(query_url, headers=headers)
            try:
                response_data = response.json()
                if response_data["data_status"] == "COMPLETED":
                    return response_data
                time.sleep(self.POLLING_INTERVAL)
            except (json.JSONDecodeError, KeyError) as e:
                if BIZYAIR_DEBUG:
                    print(
                        f"Response parsing error: {e} | Raw response: {response.text}"
                    )
                time.sleep(self.POLLING_INTERVAL)

        raise TimeoutError("Task processing timeout")

    def _process_result(self, result_data):
        video_url = result_data["data"]["payload"]
        return (download_url_to_video_output(video_url),)

    def generate_video(
        self,
        model_id: str,
        prompt: str,
        negative_prompt: str = "",
        seed: int = 0,
        image=None,
        lora_config=[],
        use_teacache="enable",
        **kwargs,
    ):
        if download_url_to_video_output is None or tensor_to_base64_string is None:
            raise ImportError(ERROR_MSG)

        req_dict = {}
        req_dict["guidance_scale"] = kwargs.pop("cfg", 6.0)
        req_dict["num_inference_steps"] = kwargs.pop("steps", 30)
        req_dict["prompt"] = prompt
        req_dict["negative_prompt"] = negative_prompt
        req_dict["seed"] = seed
        if lora_config:
            if len(lora_config) > 1:
                raise NotImplementedError(f"TODO, tmp only support one lora")
            req_dict["lora_name_list"] = [x[0] for x in lora_config]
            req_dict["lora_weight_list"] = [x[1] for x in lora_config]
        else:
            req_dict["lora_name_list"] = []
            req_dict["lora_weight_list"] = []

        if use_teacache == "enable":
            req_dict["teacache"] = 0.3
        else:
            req_dict["teacache"] = 0

        req_dict["image"] = self._encode_image(image_tensor=image)
        endpoint = self.MODEL_ENDPOINTS[model_id]
        query_url = self._send_initial_request(endpoint, request_data=req_dict)
        result = self._poll_for_completion(query_url)
        return self._process_result(result)



from bizyengine.core import BizyAirBaseNode


class CogView4_6B_Pipe(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(self):
        default_prompt = "A vibrant cherry red sports car sits proudly under the gleaming sun, its polished exterior smooth and flawless, casting a mirror-like reflection. The car features a low, aerodynamic body, angular headlights that gaze forward like predatory eyes, and a set of black, high-gloss racing rims that contrast starkly with the red. A subtle hint of chrome embellishes the grille and exhaust, while the tinted windows suggest a luxurious and private interior. The scene conveys a sense of speed and elegance, the car appearing as if it's about to burst into a sprint along a coastal road, with the ocean's azure waves crashing in the background.The license plate number of the car is 'CogView4'. The car sprinted along a coastal road, with the same sports car printed on the roadside billboard and large Chinese text '遥遥领先' written to it. The text was yellow, with thick strokes and heavy shadow lines."
        return {
            "required": {
                "prompt": ("STRING", {"default": default_prompt, "multiline": True}),
                "seed": ("INT", {"default": 0, "min": 0, "max": 1e14}),
                "guidance_scale": (
                    "FLOAT",
                    {"default": 3.5, "min": 0.1, "max": 100, "step": 0.1},
                ),
                "num_images_per_prompt": (
                    "INT",
                    {"default": 1, "min": 1, "max": 4, "step": 1},
                ),
                "num_inference_steps": (
                    "INT",
                    {"default": 30, "min": 1, "max": 100, "step": 1},
                ),
                "width": ("INT", {"default": 1536, "min": 16, "max": 4096, "step": 16}),
                "height": ("INT", {"default": 832, "min": 16, "max": 4096, "step": 16}),
            }
        }

    RETURN_TYPES = ("IMAGE",)
    RETURN_NAMES = ("image",)
    CATEGORY = "CogView4 Wrapper"



import base64
import json
import os
import pickle
import urllib.parse
import urllib.request
import zlib
from typing import List, Tuple, Union

import numpy as np
from bizyengine.core.common.env_var import BIZYAIR_SERVER_ADDRESS

BIZYAIR_DEBUG = os.getenv("BIZYAIR_DEBUG", False)


def send_post_request(api_url, payload, headers):
    """
    Sends a POST request to the specified API URL with the given payload and headers.

    Args:
        api_url (str): The URL of the API endpoint.
        payload (dict): The payload to send in the POST request.
        headers (dict): The headers to include in the POST request.

    Raises:
        Exception: If there is an error connecting to the server or the request fails.
    """
    try:
        data = json.dumps(payload).encode("utf-8")
        req = urllib.request.Request(api_url, data=data, headers=headers, method="POST")
        with urllib.request.urlopen(req) as response:
            response_data = response.read().decode("utf-8")
        return response_data
    except urllib.error.URLError as e:
        if "Unauthorized" in str(e):
            raise Exception(
                "Key is invalid, please refer to https://cloud.siliconflow.cn to get the API key.\n"
                "If you have the key, please click the 'BizyAir Key' button at the bottom right to set the key."
            )
        else:
            raise Exception(
                f"Failed to connect to the server: {e}, if you have no key, "
            )


def serialize_and_encode(obj: Union[np.ndarray], compress=True) -> Tuple[str, bool]:
    """
    Serializes a Python object, optionally compresses it, and then encodes it in base64.

    Args:
        obj: The Python object to serialize.
        compress (bool): Whether to compress the serialized object using zlib. Default is True.

    Returns:
        str: The base64 encoded string of the serialized (and optionally compressed) object.
    """
    serialized_obj = pickle.dumps(obj)

    if compress:
        serialized_obj = zlib.compress(serialized_obj)

    if BIZYAIR_DEBUG:
        print(
            f"serialize_and_encode: size of bytes is {format_bytes(len(serialized_obj))}"
        )

    encoded_obj = base64.b64encode(serialized_obj).decode("utf-8")

    if BIZYAIR_DEBUG:
        print(
            f"serialize_and_encode: size of base64 text is {format_bytes(len(serialized_obj))}"
        )

    return (encoded_obj, compress)


def decode_and_deserialize(response_text) -> np.ndarray:
    if BIZYAIR_DEBUG:
        print(
            f"decode_and_deserialize: size of text is {format_bytes(len(response_text))}"
        )

    ret = json.loads(response_text)

    if "result" in ret:
        msg = json.loads(ret["result"])
    else:
        msg = ret
    if msg["type"] not in (
        "comfyair",
        "bizyair",
    ):  # DO NOT CHANGE THIS LINE: "comfyair" is the type from the server node
        # TODO: change both server and client "comfyair" to "bizyair"
        raise Exception(f"Unexpected response type: {msg}")

    data = msg["data"]

    tensor_bytes = base64.b64decode(data["payload"])
    if data.get("is_compress", None):
        tensor_bytes = zlib.decompress(tensor_bytes)

    if BIZYAIR_DEBUG:
        print(
            f"decode_and_deserialize: size of bytes is {format_bytes(len(tensor_bytes))}"
        )

    deserialized_object = pickle.loads(tensor_bytes)
    return deserialized_object


def format_bytes(num_bytes: int) -> str:
    """
    Converts a number of bytes to a human-readable string with units (B, KB, or MB).

    :param num_bytes: The number of bytes to convert.
    :return: A string representing the number of bytes in a human-readable format.
    """
    if num_bytes < 1024:
        return f"{num_bytes} B"
    elif num_bytes < 1024 * 1024:
        return f"{num_bytes / 1024:.2f} KB"
    else:
        return f"{num_bytes / (1024 * 1024):.2f} MB"


def get_api_key():
    from bizyengine.core.common import get_api_key as bcc_get_api_key

    return bcc_get_api_key()


def get_llm_response(
    model: str,
    system_prompt: str,
    user_prompt: str,
    max_tokens: int = 1024,
    temperature: float = 0.7,
):
    api_url = f"{BIZYAIR_SERVER_ADDRESS}/chat/completions"
    API_KEY = get_api_key()
    headers = {
        "accept": "application/json",
        "content-type": "application/json",
        "Authorization": f"Bearer {API_KEY}",
    }

    payload = {
        "model": model,
        "messages": [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": user_prompt},
        ],
        "max_tokens": max_tokens,
        "temperature": temperature,
        "top_p": 0.9,
        "top_k": 50,
        "stream": False,
        "n": 1,
    }
    response = send_post_request(api_url, headers=headers, payload=payload)
    return response


def get_vlm_response(
    model: str,
    system_prompt: str,
    user_prompt: str,
    base64_images: List[str],
    max_tokens: int = 1024,
    temperature: float = 0.7,
    detail: str = "auto",
):
    api_url = f"{BIZYAIR_SERVER_ADDRESS}/chat/completions"
    API_KEY = get_api_key()
    headers = {
        "accept": "application/json",
        "content-type": "application/json",
        "Authorization": f"Bearer {API_KEY}",
    }

    messages = [
        {
            "role": "user",
            "content": [{"type": "text", "text": system_prompt}],
        },  # 此方法皆适用于两种 VL 模型
        # {
        #     "role": "system",
        #     "content": system_prompt,
        # },  # role 为 "system" 的这种方式只适用于 QwenVL 系列模型,并不适用于 InternVL 系列模型
    ]

    user_content = []
    for base64_image in base64_images:
        user_content.append(
            {
                "type": "image_url",
                "image_url": {
                    "url": f"data:image/webp;base64,{base64_image}",
                    "detail": detail,
                },
            }
        )
    user_content.append({"type": "text", "text": user_prompt})

    messages.append({"role": "user", "content": user_content})

    payload = {
        "model": model,
        "messages": messages,
        "max_tokens": max_tokens,
        "temperature": temperature,
        "top_p": 0.9,
        "top_k": 50,
        "stream": False,
        "n": 1,
    }

    response = send_post_request(api_url, headers=headers, payload=payload)
    return response



from bizyengine.core import BizyAirBaseNode


class TrainDatasetAdd(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "dataset_config": ("JSON",),
                "width": (
                    "INT",
                    {"min": 64, "default": 1024, "tooltip": "base resolution width"},
                ),
                "height": (
                    "INT",
                    {"min": 64, "default": 1024, "tooltip": "base resolution height"},
                ),
                "batch_size": (
                    "INT",
                    {
                        "min": 1,
                        "default": 2,
                        "tooltip": "Higher batch size uses more memory and generalizes the training more",
                    },
                ),
                "dataset_path": (
                    [
                        "to choose",
                    ],
                    # "STRING",
                    # {
                    #     "multiline": True,
                    #     "default": "",
                    #     "tooltip": "path to dataset, root is the 'ComfyUI' folder, with windows portable 'ComfyUI_windows_portable'",
                    # },
                ),
                "dataset_version_id": (
                    "STRING",
                    {
                        "default": "",
                    },
                ),
                "class_tokens": (
                    "STRING",
                    {
                        "multiline": False,
                        "default": "trigger_word",
                        "tooltip": "aka trigger word, if specified, will be added to the start of each caption, if no captions exist, will be used on it's own",
                    },
                ),
                "enable_bucket": (
                    "BOOLEAN",
                    {
                        "default": True,
                        "tooltip": "enable buckets for multi aspect ratio training",
                    },
                ),
                "bucket_no_upscale": (
                    "BOOLEAN",
                    {
                        "default": False,
                        "tooltip": "don't allow upscaling when bucketing",
                    },
                ),
                "num_repeats": (
                    "INT",
                    {
                        "default": 1,
                        "min": 1,
                        "tooltip": "number of times to repeat dataset for an epoch",
                    },
                ),
                "min_bucket_reso": (
                    "INT",
                    {
                        "default": 256,
                        "min": 64,
                        "max": 4096,
                        "step": 8,
                        "tooltip": "min bucket resolution",
                    },
                ),
                "max_bucket_reso": (
                    "INT",
                    {
                        "default": 1024,
                        "min": 64,
                        "max": 4096,
                        "step": 8,
                        "tooltip": "max bucket resolution",
                    },
                ),
            },
        }

    RETURN_TYPES = ("JSON",)
    RETURN_NAMES = ("dataset",)
    # FUNCTION = "create_config"
    CATEGORY = "FluxTrainer"



from .nodes_advanced_refluxcontrol import *
from .nodes_cogview4 import *
from .nodes_comfyui_detail_daemon import *
from .nodes_comfyui_instantid import *
from .nodes_comfyui_layerstyle_advance import *
from .nodes_comfyui_pulid_flux import *
from .nodes_controlnet import *
from .nodes_custom_sampler import *
from .nodes_dataset import *
from .nodes_differential_diffusion import *
from .nodes_flux import *
from .nodes_image_utils import *
from .nodes_ip2p import *
from .nodes_ipadapter_plus.nodes_ipadapter_plus import *
from .nodes_janus_pro import *
from .nodes_kolors_mz import *
from .nodes_model_advanced import *
from .nodes_nunchaku import *
from .nodes_sd3 import *
from .nodes_segment_anything import *
from .nodes_testing_utils import *
from .nodes_trellis import *
from .nodes_ultimatesdupscale import *
from .nodes_upscale_model import *
from .nodes_wan_i2v import *
from .nodes_wan_video import *
from .route_bizyair_tools import *



class ErrorNo:
    def __init__(self, http_status_code, code, payload, messages):
        self.http_status_code = http_status_code
        self.code = code
        self.messages = messages  # 现在是一个字典，key是语言代码
        self.data = payload

    def copy(self):
        return ErrorNo(self.http_status_code, self.code, self.data, self.messages)

    def get_message(self, lang="en"):
        return self.messages.get(lang, self.messages.get("en", "Unknown error"))


class errnos:
    OK = ErrorNo(200, 20000, None, {"en": "Success", "zh": "成功"})
    NO_MODEL_FOUND = ErrorNo(
        404, 20226, None, {"en": "No model found", "zh": "未找到模型"}
    )
    INTERNAL_ERROR = ErrorNo(
        500, 50000, None, {"en": "Internal Error", "zh": "内部错误"}
    )

    INVALID_TYPE = ErrorNo(
        400, 400100, None, {"en": "Invalid model type", "zh": "无效的模型类型"}
    )
    INVALID_NAME = ErrorNo(
        400, 400101, None, {"en": "Invalid model name", "zh": "无效的模型名称"}
    )
    NO_FILE_UPLOAD = ErrorNo(
        400, 400102, None, {"en": "No file to upload", "zh": "没有要上传的文件"}
    )
    EMPTY_UPLOAD_ID = ErrorNo(
        400, 400103, None, {"en": "Upload id is empty", "zh": "上传ID为空"}
    )
    INVALID_SHARE_ID = ErrorNo(
        400, 400104, None, {"en": "Invalid share id", "zh": "无效的分享ID"}
    )
    EMPTY_ABS_PATH = ErrorNo(
        400,
        400111,
        None,
        {"en": "The upload path cannot be empty", "zh": "上传路径不能为空"},
    )
    NO_ABS_PATH = ErrorNo(
        400,
        400112,
        None,
        {"en": "The upload path is not an absolute path", "zh": "上传路径不是绝对路径"},
    )
    PATH_NOT_EXISTS = ErrorNo(
        400,
        400113,
        None,
        {"en": "The upload path does not exist", "zh": "上传路径不存在"},
    )
    INVALID_CLIENT_ID = ErrorNo(
        400, 400114, None, {"en": "Invalid client id", "zh": "无效的客户端ID"}
    )
    NO_PUBLIC_FLAG = ErrorNo(
        400,
        400115,
        None,
        {"en": 'The parameter "public" is not provided', "zh": "未提供public参数"},
    )
    FILE_NOT_EXISTS = ErrorNo(
        400, 400116, None, {"en": "The file does not exist", "zh": "文件不存在"}
    )
    NO_SHARE_ID = ErrorNo(
        400,
        400117,
        None,
        {"en": 'The parameter "share_id" is not provided', "zh": "未提供share_id参数"},
    )
    INVALID_DESCRIPTION = ErrorNo(
        400, 400118, None, {"en": "Invalid description", "zh": "无效的描述"}
    )
    NOT_A_FILE = ErrorNo(
        400, 400119, None, {"en": "The path is not a file", "zh": "路径不是文件"}
    )
    NOT_ALLOWED_EXT_NAME = ErrorNo(
        400,
        400120,
        None,
        {"en": "Not allowed extension name", "zh": "不允许的文件扩展名"},
    )
    INVALID_UPLOAD_ID = ErrorNo(
        400, 400121, None, {"en": "Invalid upload id", "zh": "无效的上传ID"}
    )
    INVALID_VERSIONS = ErrorNo(
        400, 400122, None, {"en": "Invalid versions", "zh": "无效的版本信息"}
    )
    DUPLICATE_VERSION = ErrorNo(
        400, 400123, None, {"en": "Duplicate version", "zh": "重复的版本"}
    )
    INVALID_VERSION_FIELD = ErrorNo(
        400, 400124, None, {"en": "Invalid version field", "zh": "无效的版本字段"}
    )
    INVALID_QUERY_MODE = ErrorNo(
        400, 400125, None, {"en": "Invalid query mode", "zh": "无效的查询模式"}
    )
    INVALID_VERSION_NAME = ErrorNo(
        400,
        400126,
        None,
        {"en": "Invalid model version name", "zh": "无效的模型版本名称"},
    )
    INVALID_MODEL_ID = ErrorNo(
        400, 400127, None, {"en": "Invalid model id", "zh": "无效的模型ID"}
    )
    INVALID_MODEL_VERSION_ID = ErrorNo(
        400, 400128, None, {"en": "Invalid model version id", "zh": "无效的模型版本ID"}
    )
    UNSUPPORT_LIKE_TYPE = ErrorNo(
        400, 400130, None, {"en": "Unsupport like type", "zh": "不支持的点赞类型"}
    )
    INVALID_SIGN = ErrorNo(
        400, 400131, None, {"en": "Invalid file sign", "zh": "无效的文件签名"}
    )
    EMPTY_SHA256SUM = ErrorNo(
        400, 400132, None, {"en": "Empty sha256sum", "zh": "空的sha256sum"}
    )
    INVALID_OBJECT_KEY = ErrorNo(
        400, 400133, None, {"en": "Invalid object key", "zh": "无效的对象键"}
    )
    FAILED_TO_FETCH_WORKFLOW_JSON = ErrorNo(
        400,
        400134,
        None,
        {"en": "Failed to fetch workflow json", "zh": "获取工作流JSON失败"},
    )
    INVALID_DATASET_NAME = ErrorNo(
        400, 400135, None, {"en": "Invalid dataset name", "zh": "无效的数据集名称"}
    )
    INVALID_DATASET_VERSION = ErrorNo(
        400, 400136, None, {"en": "Invalid dataset version", "zh": "无效的数据集版本"}
    )
    INVALID_DATASET_ID = ErrorNo(
        400, 400137, None, {"en": "Invalid dataset id", "zh": "无效的数据集ID"}
    )
    INVALID_DATASET_VERSION_ID = ErrorNo(
        400,
        400138,
        None,
        {"en": "Invalid dataset version id", "zh": "无效的数据集版本ID"},
    )
    INVALID_SHARE_BIZ_ID = ErrorNo(
        400, 400139, None, {"en": "Invalid share biz id", "zh": "无效的分享业务ID"}
    )
    INVALID_SHARE_TYPE = ErrorNo(
        400, 400140, None, {"en": "Invalid share type", "zh": "无效的分享类型"}
    )
    INVALID_SHARE_CODE = ErrorNo(
        400, 400141, None, {"en": "Invalid share code", "zh": "无效的分享代码"}
    )
    INVALID_NOTIF_ID = ErrorNo(
        400, 400142, None, {"en": "Invalid notification id", "zh": "无效的通知ID"}
    )
    INVALID_PRODUCT_ID = ErrorNo(
        400, 400143, None, {"en": "Invalid product id", "zh": "无效的产品ID"}
    )
    INVALID_ORDER_NO = ErrorNo(
        400, 400144, None, {"en": "Invalid order number", "zh": "无效的订单号"}
    )

    INVALID_PAY_PLATFORM = ErrorNo(
        400, 400145, None, {"en": "Invalid pay platform", "zh": "无效的支付平台"}
    )

    INVALID_YEAR_PARAM = ErrorNo(
        400, 400146, None, {"en": "Year parameter is required", "zh": "缺少年份参数"}
    )

    INVALID_MONTH_PARAM = ErrorNo(
        400, 400147, None, {"en": "Month parameter is required", "zh": "缺少月份参数"}
    )

    INVALID_DAY_PARAM = ErrorNo(
        400, 400148, None, {"en": "Day parameter is required", "zh": "缺少日期参数"}
    )

    INVALID_API_KEY = ErrorNo(
        401, 401000, None, {"en": "Invalid API key", "zh": "无效的API密钥"}
    )
    INVALID_USER = ErrorNo(
        401, 401001, None, {"en": "Invalid user", "zh": "无效的用户"}
    )

    SIGN_FILE = ErrorNo(
        500, 500101, None, {"en": "Failed to sign file", "zh": "文件签名失败"}
    )
    UPLOAD = ErrorNo(
        500, 500102, None, {"en": "Failed to upload file", "zh": "文件上传失败"}
    )
    COMMIT_FILE = ErrorNo(
        500, 500103, None, {"en": "Failed to commit file", "zh": "文件提交失败"}
    )
    COMMIT_BIZY_MODEL = ErrorNo(
        500, 500105, None, {"en": "Failed to commit model", "zh": "模型提交失败"}
    )
    EMPTY_FILES = ErrorNo(
        500,
        500107,
        None,
        {"en": "Empty files to make a model", "zh": "创建模型的文件为空"},
    )
    LIST_MODEL_FILE = ErrorNo(
        500, 500108, None, {"en": "Failed to list model file", "zh": "列出模型文件失败"}
    )
    LIST_SHARE_MODEL_FILE = ErrorNo(
        500,
        500108,
        None,
        {"en": "Failed to list share model file", "zh": "列出分享模型文件失败"},
    )
    INVALID_FILENAME = ErrorNo(
        500, 500109, None, {"en": "Invalid filename", "zh": "无效的文件名"}
    )
    DELETE_MODEL = ErrorNo(
        500, 500110, None, {"en": "Failed to delete model", "zh": "删除模型失败"}
    )
    GET_USER_INFO = ErrorNo(
        500, 500111, None, {"en": "Failed to get user info", "zh": "获取用户信息失败"}
    )
    LIST_MODEL = ErrorNo(
        500, 500112, None, {"en": "Failed to list model", "zh": "列出模型失败"}
    )
    CHANGE_PUBLIC = ErrorNo(
        500, 500113, None, {"en": "Failed to change public", "zh": "更改公开状态失败"}
    )
    UPDATE_SHARE_ID = ErrorNo(
        500, 500114, None, {"en": "Failed to update share id", "zh": "更新分享ID失败"}
    )
    GET_DESCRIPTION = ErrorNo(
        500, 500115, None, {"en": "Failed to get description", "zh": "获取描述失败"}
    )
    UPDATE_DESCRIPTION = ErrorNo(
        500, 500116, None, {"en": "Failed to update description", "zh": "更新描述失败"}
    )
    DELETE_BIZY_MODEL = ErrorNo(
        500, 500117, None, {"en": "Failed to delete model", "zh": "删除模型失败"}
    )
    QUERY_COMMUNITY_MODELS = ErrorNo(
        500,
        500118,
        None,
        {"en": "Failed to query community models", "zh": "查询社区模型失败"},
    )
    QUERY_MODELS = ErrorNo(
        500, 500119, None, {"en": "Failed to query models", "zh": "查询模型失败"}
    )
    GET_MODEL_DETAIL = ErrorNo(
        500,
        500120,
        None,
        {"en": "Failed to get model detail", "zh": "获取模型详情失败"},
    )
    GET_MODEL_VERSION_DETAIL = ErrorNo(
        500,
        500121,
        None,
        {"en": "Failed to get model version detail", "zh": "获取模型版本详情失败"},
    )
    FORK_MODEL_VERSION = ErrorNo(
        500,
        500122,
        None,
        {"en": "Failed to fork model version", "zh": "分支模型版本失败"},
    )
    UPDATE_MODEL = ErrorNo(
        500, 500123, None, {"en": "Failed to update model", "zh": "更新模型失败"}
    )
    GET_UPLOAD_TOKEN = ErrorNo(
        500,
        500124,
        None,
        {"en": "Failed to get upload token", "zh": "获取上传令牌失败"},
    )
    TOGGLE_USER_LIKE = ErrorNo(
        500,
        500125,
        None,
        {"en": "Failed to toggle user like", "zh": "切换用户点赞状态失败"},
    )
    GET_DOWNLOAD_URL = ErrorNo(
        500, 500126, None, {"en": "Failed to get download url", "zh": "获取下载URL失败"}
    )
    DOWNLOAD_JSON = ErrorNo(
        500, 500127, None, {"en": "Failed to download json", "zh": "下载JSON失败"}
    )
    LIST_SHARE_MODEL_FILE_ERR = ErrorNo(
        500,
        500128,
        None,
        {"en": "Failed to list share model file", "zh": "列出分享模型文件失败"},
    )
    QUERY_OFFICIAL_MODELS = ErrorNo(
        500,
        500129,
        None,
        {"en": "Failed to query official models", "zh": "查询官方模型失败"},
    )
    UNFORK_MODEL_VERSION = ErrorNo(
        500,
        500130,
        None,
        {"en": "Failed to unfork model version", "zh": "取消分支模型版本失败"},
    )

    COMMIT_DATASET = ErrorNo(
        500, 500130, None, {"en": "Failed to commit dataset", "zh": "提交数据集失败"}
    )
    UPDATE_DATASET = ErrorNo(
        500, 500131, None, {"en": "Failed to update dataset", "zh": "更新数据集失败"}
    )
    GET_DATASET_VERSION_DETAIL = ErrorNo(
        500,
        500132,
        None,
        {"en": "Failed to get dataset version detail", "zh": "获取数据集版本详情失败"},
    )
    DELETE_DATASET = ErrorNo(
        500, 500133, None, {"en": "Failed to delete dataset", "zh": "删除数据集失败"}
    )
    QUERY_DATASETS = ErrorNo(
        500, 500134, None, {"en": "Failed to query datasets", "zh": "查询数据集失败"}
    )
    GET_DATASET_DETAIL = ErrorNo(
        500,
        500135,
        None,
        {"en": "Failed to get dataset detail", "zh": "获取数据集详情失败"},
    )
    CREATE_SHARE = ErrorNo(
        500, 500136, None, {"en": "Failed to create share", "zh": "创建分享失败"}
    )
    GET_SHARE_DETAIL = ErrorNo(
        500,
        500137,
        None,
        {"en": "Failed to get share detail", "zh": "获取分享详情失败"},
    )

    GET_DATA_DICT = ErrorNo(
        500, 500138, None, {"en": "Failed to get data dict", "zh": "获取数据字典失败"}
    )

    GET_NOTIF_UNREAD_COUNT = ErrorNo(
        500,
        500139,
        None,
        {
            "en": "Failed to get notification unread counts",
            "zh": "获取未读通知数量失败",
        },
    )
    QUERY_NOTIF = ErrorNo(
        500, 500140, None, {"en": "Failed to query notifications", "zh": "查询通知失败"}
    )
    READ_NOTIF = ErrorNo(
        500,
        500141,
        None,
        {"en": "Failed to mark notifications as read", "zh": "标记通知为已读失败"},
    )
    READ_ALL_NOTIF = ErrorNo(
        500,
        500142,
        None,
        {
            "en": "Failed to mark all notifications as read",
            "zh": "标记所有通知为已读失败",
        },
    )
    GET_WALLET_INFO = ErrorNo(
        500,
        500143,
        None,
        {"en": "Failed to get user wallet info", "zh": "获取用户钱包信息失败"},
    )
    QUERY_COINS = ErrorNo(
        500,
        500144,
        None,
        {"en": "Failed to query user coin records", "zh": "查询用户金币记录失败"},
    )
    GET_USER_METADATA = ErrorNo(
        500,
        500145,
        None,
        {"en": "Failed to get user metadata", "zh": "获取用户元数据失败"},
    )
    UPDATE_USER_INFO = ErrorNo(
        500,
        500146,
        None,
        {"en": "Failed to update user info", "zh": "更新用户信息失败"},
    )
    USER_REAL_NAME = ErrorNo(
        500, 500147, None, {"en": "Failed to verify real name", "zh": "实名认证失败"}
    )
    COPY_PROFILE_FAILED = ErrorNo(
        500, 500148, None, {"en": "Failed to copy profile", "zh": "复制配置文件失败"}
    )
    CREATE_PROFILE_FAILED = ErrorNo(
        500, 500149, None, {"en": "Failed to create profile", "zh": "创建配置文件失败"}
    )
    WRITE_PROFILE_FAILED = ErrorNo(
        500, 500150, None, {"en": "Failed to write profile", "zh": "写入配置文件失败"}
    )
    READ_PROFILE_FAILED = ErrorNo(
        500, 500151, None, {"en": "Failed to read profile", "zh": "读取配置文件失败"}
    )
    BUY_PRODUCT = ErrorNo(
        500, 500148, None, {"en": "Failed to buy product", "zh": "购买商品失败"}
    )
    PAY_STATUS = ErrorNo(
        500,
        500149,
        None,
        {"en": "Failed to check payment status", "zh": "检查支付状态失败"},
    )
    PAY_CANCEL = ErrorNo(
        500, 500150, None, {"en": "Failed to cancel payment", "zh": "取消支付失败"}
    )
    LIST_PRODUCTS = ErrorNo(
        500, 500151, None, {"en": "Failed to list products", "zh": "列出商品失败"}
    )
    LIST_PAY_ORDER = ErrorNo(
        500, 500152, None, {"en": "Failed to list pay orders", "zh": "列出支付订单失败"}
    )
    GET_YEAR_COST = ErrorNo(
        500, 500153, None, {"en": "Failed to get year cost", "zh": "获取年度费用失败"}
    )
    GET_MONTH_COST = ErrorNo(
        500, 500154, None, {"en": "Failed to get month cost", "zh": "获取月度费用失败"}
    )
    GET_DAY_COST = ErrorNo(
        500, 500155, None, {"en": "Failed to get day cost", "zh": "获取日费用失败"}
    )

    GET_RECENT_COST = ErrorNo(
        500, 500157, None, {"en": "Failed to get recent cost", "zh": "获取最近消费失败"}
    )

    MODEL_API_ERROR = ErrorNo(
        500, 500158, None, {"en": "Failed to call model API", "zh": "调用模型API失败"}
    )

    MODEL_API_TIMEOUT = ErrorNo(
        500, 500159, None, {"en": "Model API request timeout", "zh": "模型API请求超时"}
    )

    STREAMING_CONNECTION_ERROR = ErrorNo(
        500,
        500160,
        None,
        {"en": "Streaming connection closed unexpectedly", "zh": "流式连接异常关闭"},
    )



from .path_manager import (
    convert_prompt_label_path_to_real_path,
    disable_refresh_options,
    enable_refresh_options,
    get_filename_list,
    guess_config,
    guess_url_from_node,
)
from .utils import filter_files_extensions



from .execution import upload_worker
from .server import BizyAirServer

bizy_server = BizyAirServer()

if bizy_server:
    print("\n\n\033[92m[BizyAir]\033[0m Model hosting service initialized.\n\n")



"""ComfyUI Advanced Redux Control
Reference: https://github.com/kaibioinfo/ComfyUI_AdvancedRefluxControl
"""

from bizyengine.core import BizyAirBaseNode, data_types
from bizyengine.core.path_utils import path_manager as folder_paths

STRENGTHS = ["highest", "high", "medium", "low", "lowest"]
STRENGTHS_VALUES = [1, 2, 3, 4, 5]
IMAGE_MODES = ["center crop (square)", "keep aspect ratio", "autocrop with mask"]


class StyleModelApplySimple(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning": (data_types.CONDITIONING,),
                "style_model": (data_types.STYLE_MODEL,),
                "clip_vision_output": ("CLIP_VISION_OUTPUT",),
                "image_strength": (STRENGTHS, {"default": "medium"}),
            }
        }

    RETURN_TYPES = (data_types.CONDITIONING,)
    CATEGORY = "conditioning/style_model"
    NODE_DISPLAY_NAME = "Apply style model (simple)"


class ReduxAdvanced(BizyAirBaseNode):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conditioning": (data_types.CONDITIONING,),
                "style_model": (data_types.STYLE_MODEL,),
                "clip_vision": ("CLIP_VISION",),
                "image": ("IMAGE",),
                "downsampling_factor": ("INT", {"default": 3, "min": 1, "max": 9}),
                "downsampling_function": (
                    ["nearest", "bilinear", "bicubic", "area", "nearest-exact"],
                    {"default": "area"},
                ),
                "mode": (IMAGE_MODES, {"default": "center crop (square)"}),
                "weight": (
                    "FLOAT",
                    {"default": 1.0, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
            },
            "optional": {
                "mask": ("MASK",),
                "autocrop_margin": (
                    "FLOAT",
                    {"default": 0.1, "min": 0.0, "max": 1.0, "step": 0.01},
                ),
            },
        }

    RETURN_TYPES = (data_types.CONDITIONING, "IMAGE", "MASK")
    # FUNCTION = "apply_stylemodel"
    NODE_DISPLAY_NAME = "Apply Redux model (advanced)"
    CATEGORY = "conditioning/style_model"
